from __future__ import annotations

import os
from pathlib import Path

import numpy
import pint
import pytest
from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill import converter
from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.tests.datasets import GitlabDataset

_ureg = pint.get_application_registry()


@pytest.fixture(scope="class")
def S59_68p41b_16mm_6p5mm_F8_0001_Dataset() -> Path:
    scan_dir = GitlabDataset.get_dataset("h5_datasets")
    assert scan_dir is not None
    return Path(os.path.join(scan_dir, "legacy", "S59_68p41b_16mm_6p5mm_F8_0001.h5"))


@pytest.fixture(scope="class")
def WGN_01_0000_P_110_8128_D_129_Dataset():
    scan_dir = GitlabDataset.get_dataset("h5_datasets")
    assert scan_dir is not None
    return Path(os.path.join(scan_dir, "legacy", "WGN_01_0000_P_110_8128_D_129.h5"))


def test_S59_68p41b_16mm_6p5mm_F8_0001_dataset(
    S59_68p41b_16mm_6p5mm_F8_0001_Dataset, tmp_path
):
    """test 'S59_68p41b_16mm_6p5mm_F8_0001.h5' legacy dataset"""
    output_file = tmp_path / "S59_68p41b_16mm_6p5mm_F8_0001.nx"
    assert not os.path.exists(output_file), "output_file exists already"
    configuration = TomoHDF5Config()

    configuration.output_file = str(output_file)
    configuration.input_file = str(S59_68p41b_16mm_6p5mm_F8_0001_Dataset)
    configuration.raises_error = False
    configuration.no_master_file = False
    configuration.single_file = True

    result = converter.from_h5_to_nx(configuration=configuration)
    assert len(result) == 1

    assert os.path.exists(output_file)

    nxtomo = NXtomo().load(*result[0])

    nxtomo.energy = 70.0 * _ureg.keV
    n_frames = 10252

    assert len(nxtomo.sample.x_translation) == n_frames
    assert len(nxtomo.sample.y_translation) == n_frames
    assert len(nxtomo.sample.z_translation) == n_frames
    assert len(nxtomo.instrument.detector.count_time) == n_frames
    assert len(nxtomo.instrument.detector.image_key) == n_frames
    assert len(nxtomo.instrument.detector.image_key_control) == n_frames
    assert nxtomo.instrument.detector.field_of_view.value == "Full"

    N_PROJ = 10000

    # check image keys
    expected_image_key = numpy.concatenate(
        [
            [ImageKey.DARK_FIELD] * 50,
            [ImageKey.FLAT_FIELD] * 101,
            [ImageKey.PROJECTION] * N_PROJ,
            [ImageKey.FLAT_FIELD] * 101,
        ]
    )
    numpy.testing.assert_array_equal(
        nxtomo.instrument.detector.image_key_control, expected_image_key
    )

    # check angles
    expected_angles = numpy.concatenate(
        [
            [0] * 151,
            numpy.linspace(0, 360, N_PROJ, endpoint=False),
            [0] * 101,
        ]
    )
    assert len(expected_angles) == n_frames
    numpy.testing.assert_almost_equal(
        nxtomo.sample.rotation_angle.to(_ureg.degree).magnitude,
        expected_angles,
        decimal=1,
    )

    assert nxtomo.instrument.detector.distance == 105 * _ureg.millimeter
    assert nxtomo.instrument.source.distance == -145000 * _ureg.millimeter

    assert nxtomo.instrument.detector.x_pixel_size == 6.5 * _ureg.micrometer
    assert nxtomo.instrument.detector.y_pixel_size == 6.5 * _ureg.micrometer


def test_WGN_01_0000_P_110_8128_D_129(WGN_01_0000_P_110_8128_D_129_Dataset, tmp_path):
    """test 'WGN_01_0000_P_110_8128_D_129' legacy dataset"""
    output_file = tmp_path / "WGN_01_0000_P_110_8128_D_129_Dataset.nx"
    assert not os.path.exists(output_file), "output_file exists already"
    configuration = TomoHDF5Config()

    configuration.output_file = str(output_file)
    configuration.input_file = str(WGN_01_0000_P_110_8128_D_129_Dataset)
    configuration.raises_error = False
    configuration.no_master_file = False
    configuration.single_file = True

    result = converter.from_h5_to_nx(configuration=configuration)
    assert len(result) == 3

    assert os.path.exists(output_file)

    nxtomo = NXtomo().load(*result[0])

    nxtomo.energy = 80.0 * _ureg.keV
    n_frames = 4602

    assert len(nxtomo.sample.x_translation) == n_frames
    assert len(nxtomo.sample.y_translation) == n_frames
    assert len(nxtomo.sample.z_translation) == n_frames
    assert len(nxtomo.instrument.detector.count_time) == n_frames
    assert len(nxtomo.instrument.detector.image_key) == n_frames
    assert len(nxtomo.instrument.detector.image_key_control) == n_frames
    assert nxtomo.instrument.detector.field_of_view.value == "Full"

    N_PROJ = 4500

    # check image keys
    expected_image_key = numpy.concatenate(
        [
            [ImageKey.DARK_FIELD] * 100,
            [ImageKey.FLAT_FIELD] * 2,
            [ImageKey.PROJECTION] * N_PROJ,
        ]
    )
    numpy.testing.assert_array_equal(
        nxtomo.instrument.detector.image_key_control, expected_image_key
    )

    # check angles
    expected_angles = numpy.concatenate(
        [
            [0] * 102,
            numpy.linspace(0, 180, N_PROJ, endpoint=False),
        ]
    )
    assert len(expected_angles) == n_frames
    numpy.testing.assert_almost_equal(
        nxtomo.sample.rotation_angle.to(_ureg.degree).magnitude,
        expected_angles,
        decimal=1,
    )

    # note: saved value is indeed 0...
    assert nxtomo.instrument.detector.distance == 0 * _ureg.millimeter
    assert nxtomo.instrument.source.distance == -20000 * _ureg.millimeter

    assert nxtomo.instrument.detector.x_pixel_size == 6.5 * _ureg.micrometer
    assert nxtomo.instrument.detector.y_pixel_size == 6.5 * _ureg.micrometer



from __future__ import annotations

import numpy
import h5py
from silx.io.utils import h5py_read_dataset
from nxtomomill.models.utils import convert_str_to_tuple, convert_str_to_bool

__all__ = [
    "TomoConfig",
]


class TomoConfig:
    """
    hold motor used for tomography acquisition according to https://tomo.gitlab-pages.esrf.fr/bliss-tomo/master/modelization_sample_stage.html convension
    """

    def __init__(self) -> None:
        self._rotation: None | tuple[str, ...] = None
        self._sample_u: None | tuple[str, ...] = None
        self._sample_v: None | tuple[str, ...] = None
        self._sample_x: None | tuple[str, ...] = None
        self._sample_y: None | tuple[str, ...] = None
        self._translation_x: None | tuple[str, ...] = None
        self._translation_y: None | tuple[str, ...] = None
        self._translation_z: None | tuple[str, ...] = None
        self._tomo_detector: None | tuple[str, ...] = None
        self._rotation_is_clockwise: None | bool = None

    @property
    def rotation(self) -> tuple[str, ...] | None:
        return self._rotation

    @rotation.setter
    def rotation(self, motor: tuple[str, ...] | None):
        self._rotation = motor

    @property
    def sample_u(self) -> tuple[str, ...] | None:
        return self._sample_u

    @sample_u.setter
    def sample_u(self, motor: tuple[str, ...] | None):
        self._sample_u = motor

    @property
    def sample_v(self) -> tuple[str, ...] | None:
        return self._sample_v

    @sample_v.setter
    def sample_v(self, motor: tuple[str, ...] | None):
        self._sample_v = motor

    @property
    def sample_x(self) -> tuple[str, ...] | None:
        return self._sample_x

    @sample_x.setter
    def sample_x(self, motor: tuple[str, ...] | None):
        self._sample_x = motor

    @property
    def sample_y(self) -> tuple[str, ...] | None:
        return self._sample_y

    @sample_y.setter
    def sample_y(self, motor: tuple[str, ...] | None):
        self._sample_y = motor

    @property
    def translation_x(self) -> tuple[str, ...] | None:
        return self._translation_x

    @translation_x.setter
    def translation_x(self, motor: tuple[str, ...] | None):
        self._translation_x = motor

    @property
    def translation_y(self) -> tuple[str, ...] | None:
        return self._translation_y

    @translation_y.setter
    def translation_y(self, motor: tuple[str, ...] | None):
        self._translation_y = motor

    @property
    def translation_z(self) -> tuple[str, ...] | None:
        return self._translation_z

    @translation_z.setter
    def translation_z(self, motor: tuple[str, ...] | None):
        self._translation_z = motor

    @property
    def tomo_detector(self) -> tuple[str, ...] | None:
        return self._tomo_detector

    @tomo_detector.setter
    def tomo_detector(self, detector_name: tuple[str, ...]):
        self._tomo_detector = detector_name

    @property
    def rotation_is_clockwise(self) -> bool | None:
        return self._rotation_is_clockwise

    @rotation_is_clockwise.setter
    def rotation_is_clockwise(self, is_clockwise: bool | None) -> None:
        if is_clockwise is None:
            self._rotation_is_clockwise = is_clockwise
        elif isinstance(is_clockwise, (bool, numpy.bool_)):
            self._rotation_is_clockwise = bool(is_clockwise)
        else:
            raise TypeError(
                f"'is_clockwise' should be a bool or Not. Got {type(is_clockwise)}"
            )

    def __str__(self) -> str:
        return "tomo_config:" + " ; ".join(
            [
                f"rotation={', '.join(self.rotation)}",
                f"rotation_is_clockwise={self.rotation_is_clockwise}",
                f"sample_u={', '.join(self.sample_u)}",
                f"sample_v={', '.join(self.sample_v)}",
                f"sample_x={', '.join(self.sample_x)}",
                f"sample_y={', '.join(self.sample_y)}",
                f"tomo_detector={', '.join(self.tomo_detector)}",
                f"translation_x={', '.join(self.translation_x)}",
                f"translation_y={', '.join(self.translation_y)}",
                f"translation_z={', '.join(self.translation_z)}",
            ]
        )

    @staticmethod
    def from_technique_group(technique_group: h5py.Group):
        """
        get rotation motor and thinks like this from the 'tomoconfig'.
        This can retrieve one or several dataset name or a single one.
        In the case of several dataset name we get (real_motor_name, bliss_alias)

        If the motor moves then this is pretty simple the real motor_name dataset exists.
        But if the motors does not move during the bliss scan (scalar value) then the real_motor_name dataset doesn't exists and
        the bliss alias does. This is why we need to keep both and check both during the 'standard process'...
        """
        if not isinstance(technique_group, h5py.Group):
            raise TypeError(
                f"instrument_group is expected to be an instance of {h5py.Group}. {type(technique_group)} provided"
            )
        if "tomoconfig" not in technique_group:
            raise KeyError("could find 'tomoconfig' key")
        else:
            tomo_config_group = technique_group.get("tomoconfig")

        def get_dataset(group, dataset_name, default, dtype=tuple):
            if dataset_name not in group:
                return default
            dataset = h5py_read_dataset(group[dataset_name])
            # note: the dataset are read as string '["detector"]' and must be converted to dtype
            if dtype is tuple:
                return convert_str_to_tuple(tuple(dataset))
            elif dtype is bool:
                return convert_str_to_bool(dataset)
            else:
                raise TypeError(f"invalid dtype: {dtype}")

        tomo_config = TomoConfig()
        tomo_config.rotation = get_dataset(tomo_config_group, "rotation", None)
        tomo_config.rotation_is_clockwise = get_dataset(
            tomo_config_group, "rotation_is_clockwise", default=None, dtype=bool
        )
        tomo_config.sample_u = get_dataset(tomo_config_group, "sample_u", None)
        tomo_config.sample_v = get_dataset(tomo_config_group, "sample_v", None)
        tomo_config.sample_x = get_dataset(tomo_config_group, "sample_x", None)
        tomo_config.sample_y = get_dataset(tomo_config_group, "sample_y", None)
        tomo_config.tomo_detector = get_dataset(tomo_config_group, "detector", None)

        # translation x == sample_u
        tomo_config.translation_x = get_dataset(
            tomo_config_group, "sample_u", None
        )  # Needs translation X mapping
        tomo_config.translation_y = get_dataset(
            tomo_config_group, "translation_y", None
        )
        tomo_config.translation_z = get_dataset(
            tomo_config_group, "translation_z", None
        )

        return tomo_config



# coding: utf-8

"""
Application to create a default configuration file to be used by h52nx application.

.. program-output:: nxtomomill h5-config --help

For a complete tutorial you can have a look at: :ref:`Tomoh52nx`
"""

import argparse
import logging

from nxtomomill.io import TomoHDF5Config

logging.basicConfig(level=logging.INFO)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(description="Create a default configuration file")
    parser.add_argument("output_file", help="output .cfg file")
    parser.add_argument(
        "--from-title-names",
        help="Provide minimalistic configuration to make a conversion from "
        "titles names. (FRAME TYPE section is ignored). \n"
        "Exclusive with `from-scan-urls` option",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--from-scan-urls",
        help="Provide minimalistic configuration to make a conversion from "
        "scan urls. (ENTRIES and TITLES section is ignored).\n"
        "Exclusive with `from-title-names` option",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(argv[1:])
    if options.from_title_names:
        filter_sections = ("frame_type_section",)
    elif options.from_scan_urls:
        filter_sections = ("entries_and_titles_section",)
    else:
        filter_sections = ()

    configuration = TomoHDF5Config()
    configuration.to_cfg_file(
        file_path=options.output_file, filter_sections=filter_sections
    )



# coding: utf-8

"""
Module to convert from (Bliss) .h5 to (Nexus Tomo-compliant) .nx format.
"""


class Tomo:
    class H5:
        """HDF5 settings for tomography"""

        VALID_CAMERA_NAMES: tuple[str, ...] = tuple()
        # Camera names are now deduced using the `get_nx_detectors` and
        # `guess_nx_detector` functions. Alternatively, a list of detector
        # names can be provided (supports Unix shell-style wildcards), such as:
        # ("pcolinux*", "basler", "frelon*", ...)

        ROT_ANGLE_KEYS = (
            "rotm",
            "mhsrot",
            "hsrot",
            "mrsrot",
            "hrsrot",
            "srot",
            "srot_eh2",
            "diffrz",
            "hrrz_trig",
            "rot",
        )
        """Keys used to find rotation angles."""

        TRANSLATION_Y_KEYS = ("yrot", "diffty", "hry")
        """Keys used to find the Y translation below the center of rotation."""

        TRANSLATION_Z_KEYS = ("sz", "difftz", "hrz", "pz", "ntz", "samtz", "mrsz")
        """Keys used to find the Z translation below or above the center of rotation."""

        SAMPLE_X_KEYS = ("samx", "psx", "sax", "fake_sx")
        """Keys used to find the X translation above the center of rotation (direction is independent of the rotation angle)."""

        SAMPLE_Y_KEYS = ("samy", "psv", "say", "fake_sy")
        """Keys used to find the Y translation above the center of rotation (direction is independent of the rotation angle)."""

        SAMPLE_U_KEYS = ("sau", "sx", "px", "ntx", "shtx", "hrx", "fake_su")
        """Keys used to find the U translation above the center of rotation (direction is dependent of the rotation angle)."""

        SAMPLE_V_KEYS = ("sav", "sy", "py", "nty", "shty", "hry2", "fake_sv")
        """Keys used to find the V translation above the center of rotation (direction is dependent of the rotation angle)."""

        DIODE_KEYS = ("fpico3",)
        """Keys used to store diode dataset."""

        ACQ_EXPO_TIME_KEYS = ("acq_expo_time",)
        """Keys used to store acquisition exposure time."""

        INIT_TITLES = (
            "tomo:basic",
            "tomo:fullturn",
            "tomo:fulltomo",
            "sequence_of_scans",
            "tomo:halfturn",
            "tomo:multiturn",
            "tomo:helical",
            "tomo:holotomo",
            "holotomo_distance",
        )
        """Initialization scan titles. Either the value of 'technique/scan_category' if exists else the value of 'title'"""

        ZSERIE_INIT_TITLES = ("tomo:zseries",)
        """Specific titles for z-series scans."""

        MULTITOMO_INIT_TITLES = (
            "tomo:pcotomo",
            "pcotomo",
            "tomo:multitomo",
            "multitomo",
            "multitomo:basic",
            "tomo:multiturn",
            "multiturn",
        )
        """Specific titles for multi-tomo scans (also known as PCO scans). Either the value of 'technique/scan_category' if exists else the value of 'title'"""

        PCOTOMO_INIT_TITLES = MULTITOMO_INIT_TITLES
        """Deprecated. Replaced by 'MULTITOMO_INIT_TITLES'"""

        BACK_AND_FORTH_INIT_TITLES = (
            "tomo:backandforth",
            "backandforthtomo:basic",
            "backandforth",
            "tomo:back_and_forth",
            "back_and_forth",
        )
        """Specific titles for back and forth scans. Either the value of 'technique/scan_category' if exists else the value of 'title'"""

        DARK_TITLES = ("dark images", "dark")
        """Titles to determine dark field scans if 'technique/image_key' doesn't exists."""

        FLAT_TITLES = ("flat", "reference images", "ref", "refend")
        """Titles to determine flat field scans if 'technique/image_key' doesn't exists."""

        PROJ_TITLES = ("projections", "ascan rot 0", "ascan diffrz 0 180 1600 0.1")
        """Titles to determine projection scans if 'technique/image_key' doesn't exists."""

        ALIGNMENT_TITLES = ("static images", "ascan diffrz 180 0 4 0.1")
        """Titles to determine alignment scans if 'technique/image_key' doesn't exists."""

        SAMPLE_X_PIXEL_SIZE_KEYS = ("technique/optic/sample_pixel_size",)
        """Possible paths to the pixel size along the x-axis."""

        SAMPLE_Y_PIXEL_SIZE_KEYS = ("technique/optic/sample_pixel_size",)
        """Possible paths to the pixel size along the y-axis."""

        DETECTOR_X_PIXEL_SIZE_KEYS = (
            "technique/optic/optics_pixel_size",
            "technique/scan/detector_pixel_size",
            "technique/detector/{detector_name}/pixel_size",
            "technique/detector/pixel_size",
        )

        DETECTOR_Y_PIXEL_SIZE_KEYS = (
            "technique/optic/optics_pixel_size",
            "technique/scan/detector_pixel_size",
            "technique/detector/{detector_name}/pixel_size",
            "technique/detector/pixel_size",
        )

        SAMPLE_DETECTOR_DISTANCE_KEYS = ("technique/scan/sample_detector_distance",)
        """Keys used to store the sample to detector distance."""

        SOURCE_SAMPLE_DISTANCE_KEYS = ("technique/scan/source_sample_distance",)
        """Keys used to store the source to sample distance."""

        MACHINE_CURRENT_KEYS = ("current",)
        """Keys used to store machine current values."""

    class EDF:
        """EDF settings for tomography"""

        MOTOR_POS = ("motor_pos",)
        """Keys for motor positions."""

        MOTOR_MNE = ("motor_mne",)
        """Keys for motor names (mnemonics)."""

        ROT_ANGLE = ("srot", "somega")
        """Keys used to find the rotation angle."""

        X_TRANS = ("sx",)
        """Keys used to find x translation in EDF format."""

        Y_TRANS = ("sy",)
        """Keys used to find y translation in EDF format."""

        Z_TRANS = ("sz",)
        """Keys used to find z translation in EDF format."""

        MACHINE_CURRENT = ("srcur", "srcurrent")
        """Keys used to store machine current values."""

        TO_IGNORE = ("_slice_",)
        """Fields to ignore when processing EDF files."""

        DARK_NAMES = ("darkend", "dark")
        """Names identifying dark images in EDF files."""

        REFS_NAMES = ("ref", "refHST")
        """Names identifying reference images in EDF files."""



import os
import tempfile

from nxtomomill.app.h5config import main


def test_edf_config_application():
    """test nxtomomill edf-config output_file"""

    with tempfile.TemporaryDirectory() as folder:
        output_file = os.path.join(folder, "config.cfg")
        assert not os.path.exists(output_file)
        main(["edf-config", output_file])
        assert os.path.exists(output_file)



# coding: utf-8

from __future__ import annotations

from configparser import ConfigParser
from pydantic import BaseModel, ConfigDict

from ..base.FrmFlatToNestedBase import FrmFlatToNestedBase
from ..base.NestedModelBase import NestedModelBase
from .dark_and_flat_section import DarkAndFlatSection
from .general_section import GeneralSection
from .keys_section import KeysSection
from .sample_section import SampleSection
from .unit_section import UnitSection
from .detector_section import DetectorSection
from ..base.source_section import SourceSection as _SourceSection
from ..base.instrument_section import InstrumentSection
from nxtomomill.utils.io import deprecated_warning

__all__ = ["EDF2nxModel", "generate_default_edf_config"]


class _LegacySourceSection(_SourceSection, InstrumentSection):
    """Historically the instrument name and section was melt with the source section"""

    pass


class EDF2nxModel(
    FrmFlatToNestedBase,
    GeneralSection,
    KeysSection,
    DarkAndFlatSection,
    SampleSection,
    _LegacySourceSection,
    DetectorSection,
    UnitSection,
):
    """Configuration file to run a conversion from spec-edf to NXtomo format"""

    model_config: ConfigDict = ConfigDict(validate_assignment=True)

    def to_nested_model(self) -> _NestedTomoEDFConfig:
        return _NestedTomoEDFConfig(
            general_section=GeneralSection(
                output_file=self.output_file,
                overwrite=self.overwrite,
                log_level=self.log_level,
                input_folder=self.input_folder,
                file_extension=self.file_extension,
                delete_edf_source_files=self.delete_edf_source_files,
                output_checks=self.output_checks,
                dataset_basename=self.dataset_basename,
                dataset_info_file=self.dataset_info_file,
                title=self.title,
                patterns_to_ignores=self.patterns_to_ignores,
                duplicate_data=self.duplicate_data,
                external_link_type=self.external_link_type,
            ),
            edf_keys_section=KeysSection(
                motor_position_keys=self.motor_position_keys,
                motor_mne_keys=self.motor_mne_keys,
                rotation_angle_keys=self.rotation_angle_keys,
                x_translation_keys=self.x_translation_keys,
                y_translation_keys=self.y_translation_keys,
                z_translation_keys=self.z_translation_keys,
                machine_current_keys=self.machine_current_keys,
            ),
            sample_section=SampleSection(
                angle_calculation_endpoint=self.angle_calculation_endpoint,
                force_angle_calculation=self.force_angle_calculation,
                angle_calculation_rev_neg_scan_range=self.angle_calculation_rev_neg_scan_range,
                sample_name=self.sample_name,
            ),
            source_section=_LegacySourceSection(
                instrument_name=self.instrument_name,
                source_name=self.source_name,
                source_type=self.source_type,
                source_probe=self.source_probe,
            ),
            detector_section=DetectorSection(
                field_of_view=self.field_of_view,
            ),
            unit_section=UnitSection(
                pixel_size_unit=self.pixel_size_unit,
                sample_detector_distance_unit=self.sample_detector_distance_unit,
                energy_unit=self.energy_unit,
                machine_current_unit=self.machine_current_unit,
                x_translation_unit=self.x_translation_unit,
                y_translation_unit=self.y_translation_unit,
                z_translation_unit=self.z_translation_unit,
            ),
            dark_and_flat_section=DarkAndFlatSection(
                dark_names_prefix=self.dark_names_prefix,
                flat_names_prefix=self.flat_names_prefix,
            ),
        )

    @classmethod
    def from_cfg_file(cls, file_path: str) -> EDF2nxModel:
        txt_parser = ConfigParser(allow_no_value=True)
        txt_parser.read(file_path)

        def get_section(name, default={}):
            if txt_parser.has_section(name):
                return txt_parser[name]
            else:
                return default

        return _NestedTomoEDFConfig(
            general_section=GeneralSection(**get_section("GENERAL_SECTION")),
            edf_keys_section=KeysSection(**get_section("EDF_KEYS_SECTION")),
            dark_and_flat_section=DarkAndFlatSection(
                **get_section("DARK_AND_FLAT_SECTION")
            ),
            sample_section=SampleSection(**get_section("SAMPLE_SECTION")),
            unit_section=UnitSection(**get_section("UNIT_SECTION")),
            source_section=_LegacySourceSection(**get_section("SOURCE_SECTION")),
            detector_section=DetectorSection(**get_section("DETECTOR_SECTION")),
        ).to_flatten_config()

    @staticmethod
    def from_dict(dict_: dict) -> None:
        deprecated_warning(
            type_="function",
            name="from_dict",
            reason="replaced by pydantic 'model_dump' function",
            replacement="model_dump",
            since_version="2.0",
        )

        dict_ = {key.lower(): value for key, value in dict_.items()}
        config = _NestedTomoEDFConfig(**dict_)
        return config.to_flatten_config()


class _NestedTomoEDFConfig(
    BaseModel,
    NestedModelBase,
):
    """
    Configuration class to provide to the convert from edf to nx
    """

    model_config = ConfigDict(str_to_upper=True)

    general_section: GeneralSection = GeneralSection()
    edf_keys_section: KeysSection = KeysSection()
    dark_and_flat_section: DarkAndFlatSection = DarkAndFlatSection()
    sample_section: SampleSection = SampleSection()
    source_section: _LegacySourceSection = _LegacySourceSection()
    detector_section: DetectorSection = DetectorSection()
    unit_section: UnitSection = UnitSection()

    def to_flatten_config(self) -> EDF2nxModel:
        return EDF2nxModel(
            **self.general_section.model_dump(),
            **self.edf_keys_section.model_dump(),
            **self.dark_and_flat_section.model_dump(),
            **self.sample_section.model_dump(),
            **self.source_section.model_dump(),
            **self.detector_section.model_dump(),
            **self.unit_section.model_dump(),
        )


def generate_default_edf_config() -> dict:
    """generate a default configuration for converting spec-edf to NXtomo"""
    config = EDF2nxModel()
    return {key.upper(): value for key, value in config.model_dump().items()}



from __future__ import annotations

import logging
from configparser import ConfigParser
from pydantic import BaseModel

_logger = logging.getLogger(__name__)


class NestedModelBase:

    @staticmethod
    def cast_value_to_str(value) -> str:
        if value in (None, tuple()):
            return ""
        else:
            return str(value)

    def to_cfg_file(self, file_path: str, filter_sections: tuple[str] = tuple()):
        """
        Dump the model to a cfg_file.
        In this case the model is expected to be composed of a set of sub-model.
        The sub-models are expected to only contain Fields.
        """
        if not file_path.lower().endswith((".cfg", ".config", ".conf")):
            _logger.warning("add a valid extension to the output file")
            file_path += ".cfg"
        config = ConfigParser(allow_no_value=True)
        config.optionxform = str
        for section_name in self.__class__.model_fields:  # pylint: disable=E1101
            if section_name in filter_sections:
                continue
            section_name_upper = section_name.upper()
            config.add_section(section_name_upper)
            config.set(section_name_upper, "")

            section: BaseModel = getattr(self, section_name)
            dump_model = section.model_dump(mode="python")

            for field_name, field_value in dump_model.items():
                field_info = section.model_fields[field_name]
                config.set(section_name_upper, "# " + field_info.description)
                config.set(
                    section_name_upper,
                    field_name,
                    self.cast_value_to_str(field_value),
                )

        with open(file_path, "w") as config_file:
            config.write(config_file)



from nxtomomill.utils.io import deprecated, deprecated_warning

from nxtomomill.models.utils import (
    remove_parenthesis_or_brackets as _remove_parenthesis_or_brackets,
)
from nxtomomill.models.utils import filter_str_def as _filter_str_def
from nxtomomill.models.utils import convert_str_to_tuple as _convert_str_to_tuple
from nxtomomill.models.utils import convert_str_to_bool as _convert_str_to_bool
from nxtomomill.models.utils import is_url_path as _is_url_path
from nxtomomill.models.utils import PathType as _PathType


@deprecated(
    replacement="nxtomomill.models.utils.remove_parenthesis_or_brackets",
    since_version="2.0",
    reason="moved",
)
def remove_parenthesis_or_brackets(*args, **kwargs):
    return _remove_parenthesis_or_brackets(*args, **kwargs)


@deprecated(
    replacement="nxtomomill.models.utils.filter_str_def",
    since_version="2.0",
    reason="moved",
)
def filter_str_def(*args, **kwargs):
    return _filter_str_def(*args, **kwargs)


@deprecated(
    replacement="nxtomomill.models.utils.convert_str_to_tuple",
    since_version="2.0",
    reason="moved",
)
def convert_str_to_tuple(*args, **kwargs):
    return _convert_str_to_tuple(*args, **kwargs)


@deprecated(
    replacement="nxtomomill.models.utils.convert_str_to_bool",
    since_version="2.0",
    reason="moved",
)
def convert_str_to_bool(*args, **kwargs):
    return _convert_str_to_bool(*args, **kwargs)


@deprecated(
    replacement="nxtomomill.models.utils.is_url_path",
    since_version="2.0",
    reason="moved",
)
def is_url_path(*args, **kwargs):
    return _is_url_path(*args, **kwargs)


class PathType(_PathType):
    def __init__(self, *args, **kwargs):
        deprecated_warning(
            type_="class",
            name="PathType",
            replacement="nxtomomill.models.utils.PathType",
            reason="moved to models module",
            since_version="2.0",
        )
        super().__init__(*args, **kwargs)



"""
Application to concatenate several scans; each corresponding to a z-stage, into one nexus scan for nabu-helical
"""

import argparse
import os
import re
import json
import pint

import numpy as np
from silx.io.url import DataUrl
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.scanbase import ReducedFramesInfos
from tomoscan.io import HDF5File, get_swmr_mode

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.utils import concatenate as nx_concatenate
from nxtomo.nxobject.nxdetector import ImageKey
from nxtomomill.utils.hdf5 import DatasetReader

_ureg = pint.get_application_registry()


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def get_arguments(argv):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "--filename_template",
        required=True,
        help="""The filename template. It must contain a segment equal to "X"*ndigits   which will be replaced by the stage number """,
    )
    parser.add_argument(
        "--target_file",
        required=True,
        help="The new nexus filename that we are going to create ",
    )
    parser.add_argument(
        "--entry_name",
        required=False,
        help="Optional.Output data path aka entry_name. Its default is entry0000",
        default="entry0000",
    )
    parser.add_argument(
        "--total_nstages", type=int, required=True, help="The total number of stages"
    )
    parser.add_argument(
        "--first_stage",
        type=int,
        default=0,
        required=False,
        help="Optional. Defaults to zero. The number of the first considered stage. Use this to do a smaller sequence",
    )
    parser.add_argument(
        "--last_stage",
        type=int,
        default=-1,
        required=False,
        help="Optional. Defaults to total_nstages-1. The number of the last considered stage. Use this to do a smaller sequence",
    )
    parser.add_argument(
        "--cors_file",
        required=False,
        type=str,
        help="Optional. If given, it is a txt file with a column of centers of rotation. We expect one COR per stage. They are use to set the x_translation.\n"
        "This file can be either a one column text file, or a json file. In this case the key rotation_axis_position_list must be present and give a list",
    )
    parser.add_argument(
        "--pixel_size_m",
        required=False,
        default=None,
        help="Optional. The pixel size in meters. If given it will overwrite the nexus one, in the final nexus file",
    )
    parser.add_argument(
        "--neutral_flat",
        type=str2bool,
        required=False,
        default=False,
        help="Optional. Its default is False. If true then it will set flats to arrays filled with ones and darks with zeroes",
    )
    parser.add_argument(
        "--flats_from_reduced",
        type=str2bool,
        required=False,
        default=False,
        help="Optional. Its default is False. If true then the flats will be set from the reduced flats of the original scans",
    )
    parser.add_argument(
        "--flats_from_before_after",
        type=str2bool,
        required=False,
        default=False,
        help="""Optional. Its default is False. If true set flats from the reduced flats of  "before and after" scans.\n"
        "If given then you have to provide also the names these two scans, using parameters '--scan_before' and '--scan_after'""",
    )
    parser.add_argument(
        "--scan_before",
        required=False,
        type=str,
        help="""Optional. To be used with "flats_from_before_after selected". The name of the scan before all the scans. From this the flat/dark has been extracted""",
    )
    parser.add_argument(
        "--scan_after",
        required=False,
        type=str,
        help="""Optional. To be used with "flats_from_before_after selected".  The name of the scan after all the scans. From this the flat/dark has been extracted""",
    )

    args = parser.parse_args(argv)

    sum_bool = (
        args.neutral_flat + args.flats_from_reduced + args.flats_from_before_after
    )

    if (sum_bool) not in [0, 1]:
        message = f""" Only one or none of neutral_flat, flats_from_reduced, flats_from_before_after
        options can be selected. You selected {sum_bool} of them
        """
        raise ValueError(message)

    if args.last_stage == -1:
        args.last_stage = args.total_nstages - 1

    args.nstages = args.last_stage + 1 - args.first_stage

    if args.cors_file is None:
        args.cors = np.zeros([args.nstages], "f")
    else:
        try:  # check if it is json
            with open(args.cors_file, "r") as fj:
                json_dict = json.load(fj)
        except ValueError:
            args.cors = np.loadtxt(args.cors_file)
        else:
            args.cors = json_dict["rotation_axis_position_list"]

    pattern = re.compile("[X]+")
    # X represent the variable part of the 'template'
    # for example if we want to treat scans HA_2000_sample_0000.nx, ..., HA_2000_sample_9999.nx then
    # we expect the template to be HA_2000_sample_XXXX.nx
    # warning: If the dataset base names contains several X substrings the longest ones will be taken.
    ps = pattern.findall(args.filename_template)
    ls = list(map(len, ps))
    if len(ls) < 1:
        args.name_template = args.filename_template

        if args.first_stage != args.last_stage:
            message = f" The argument for filename_template , which was '{args.filename_template}'  does not seem to contain a pattern with multiple X for the numerical part"
            raise ValueError(message)
    else:
        idx = np.argmax(ls)
        if len(ps[idx]) < 2:
            message = f""" The argument filename_template should contain  a substring  formed by at least two 'X'
            The filename_template was {args.filename_template}
            """
            raise ValueError(message)
        args.name_template = args.filename_template.replace(
            ps[idx], "{i_stage:" + "0" + str(ls[idx]) + "d}"
        )

    args.file_list = []
    for i_stage in range(args.first_stage, args.last_stage + 1):
        args.file_list.append(args.name_template.format(i_stage=i_stage))
    args.used_current = None

    if args.pixel_size_m is not None:
        args.overwrite_pixel_size = True
        args.pixel_size_m = float(args.pixel_size_m)
    else:
        args.overwrite_pixel_size = False
        with HDF5File(args.file_list[0], "r", swmr=get_swmr_mode()) as h5f:
            args.pixel_size_m = h5f[
                os.path.join(args.entry_name, "instrument", "detector", "x_pixel_size")
            ][()]
    return args


def main(argv):
    args = get_arguments(argv[1:])
    output_file = args.target_file

    dummy_frame = None
    """the dummy frame is a frame fill with ones. It will be insert between two stages to ensure series from the twos stages will be
    perceived as 'uncontiguous'. Separates flats at the end of a stage and at the beginning.
    """
    nxt_z_list = []
    npoints_list = []
    for i_stage in range(args.first_stage, args.last_stage + 1):
        filename = args.file_list[i_stage - args.first_stage]
        # step 1: load the stage
        nxt = NXtomo()
        nxt.load(filename, data_path=args.entry_name, detector_data_as="as_data_url")
        nxt_z_list.append(nxt)

        scan = NXtomoScan(filename, args.entry_name)

        npoints_list.append(len(scan.image_key_control))

        if dummy_frame is None:
            if len(nxt.instrument.detector.data) > 0:
                # create empty frame to give it to the 'invalid_nxt' NXtomo later
                with DatasetReader(nxt.instrument.detector.data[0]) as raw_data:
                    data_type = raw_data.dtype
                # FIXME: avoid keeping some file open. not clear why this is needed
                raw_data = None
                args.img_shape = (scan.dim_2, scan.dim_1)
                with HDF5File(output_file, mode="w") as h5s:
                    h5s["empty_frame"] = np.ones(
                        (1, scan.dim_2, scan.dim_1), dtype=data_type
                    )
                    dummy_frame = DataUrl(
                        file_path=output_file, data_path="empty_frame"
                    )
            else:
                message = """The first nx file must have some data"""
                raise ValueError(message)

        # step 2: insert single 'empty' frame between the two stages (aka dummy frame)
        invalid_nxt = NXtomo(args.entry_name)
        invalid_nxt.instrument.detector.image_key_control = [ImageKey.INVALID]
        invalid_nxt.energy = nxt.energy.value
        invalid_nxt.control.data = np.ones(1, "f")
        invalid_nxt.sample.rotation_angle = [0.0] * _ureg.degree
        invalid_nxt.instrument.detector.field_of_view = (
            nxt.instrument.detector.field_of_view.value
        )
        invalid_nxt.instrument.detector.x_pixel_size = (
            nxt.instrument.detector.x_pixel_size.value
        )
        invalid_nxt.instrument.detector.y_pixel_size = (
            nxt.instrument.detector.y_pixel_size.value
        )

        invalid_nxt.instrument.detector.distance = (
            nxt.instrument.detector.distance.value
        )
        invalid_nxt.sample.x_translation = np.zeros([1], "f")
        invalid_nxt.sample.y_translation = np.zeros([1], "f")
        invalid_nxt.sample.z_translation = np.zeros([1], "f")
        invalid_nxt.instrument.detector.data = (dummy_frame,)

        nxt_z_list.append(invalid_nxt)

    nx_concatenated = nx_concatenate(nxt_z_list)
    nx_concatenated.save(
        file_path=output_file, data_path=args.entry_name, overwrite=True
    )

    with HDF5File(output_file, "r+") as output_target:
        N_total = output_target[
            os.path.join(args.entry_name, "sample", "x_translation")
        ].shape[0]

        if args.neutral_flat:
            # will set flats to arrays filled with ones and darks with zeroes
            original_currents = output_target[
                os.path.join(args.entry_name, "control", "data")
            ][()]
            original_currents[:] = 1
            del output_target[os.path.join(args.entry_name, "control", "data")]
            output_target[os.path.join(args.entry_name, "control", "data")] = (
                original_currents
            )

        do_flat = (
            args.flats_from_reduced or args.flats_from_before_after or args.neutral_flat
        )

        if do_flat or (args.cors_file is not None):
            one_for_invalid = 1
            actual_pos = 0
            old_pos = actual_pos

            i_stage = args.first_stage

            if do_flat:
                darks_dictionary = {}
                flats_dictionary = {}
                darks_infos_dictionary = {"count_time": []}
                flats_infos_dictionary = {"machine_current": [], "count_time": []}
                add_dark_to_dict(
                    args, darks_dictionary, darks_infos_dictionary, i_stage, actual_pos
                )

            for i_stage in range(args.first_stage, args.last_stage + 1):
                if do_flat:
                    add_flat_to_dict(
                        args,
                        flats_dictionary,
                        flats_infos_dictionary,
                        i_stage,
                        actual_pos,
                    )

                actual_pos += npoints_list[i_stage - args.first_stage]

                if do_flat:
                    add_flat_to_dict(
                        args,
                        flats_dictionary,
                        flats_infos_dictionary,
                        i_stage,
                        actual_pos,
                        position_in_list=1,
                    )

                # if i_stage != args.last_stage:
                actual_pos += one_for_invalid

                if args.cors_file is not None:
                    output_target[
                        os.path.join(args.entry_name, "sample", "x_translation")
                    ][old_pos:actual_pos] = (
                        -args.cors[i_stage - args.first_stage] + args.cors[0]
                    ) * args.pixel_size_m

                old_pos = actual_pos

            assert actual_pos == N_total

            if args.overwrite_pixel_size:
                g = output_target[
                    os.path.join(args.entry_name, "instrument", "detector")
                ]
                g["x_pixel_size"][()] = args.pixel_size_m
                g["y_pixel_size"][()] = args.pixel_size_m

            if do_flat:
                # darks metadata
                meta_darks = ReducedFramesInfos()
                meta_darks.count_time.extend(darks_infos_dictionary["count_time"][:1])

                # flats metadata
                meta_flats = ReducedFramesInfos()
                meta_flats.machine_current.extend(
                    flats_infos_dictionary["machine_current"]
                )
                meta_flats.count_time.extend(flats_infos_dictionary["count_time"])

                # save metadata
                scan = NXtomoScan(output_file, args.entry_name)
                scan.save_reduced_flats(
                    flats_dictionary, flats_infos=meta_flats, overwrite=True
                )
                scan.save_reduced_darks(
                    darks_dictionary, darks_infos=meta_darks, overwrite=True
                )


def add_flat_to_dict(
    args,
    flats_dictionary,
    flats_infos_dictionary,
    i_stage,
    actual_pos,
    position_in_list=0,
):
    if args.flats_from_reduced:
        filename = args.file_list[i_stage - args.first_stage]
        scan = NXtomoScan(filename, args.entry_name)
        reduced_flats, metadata_flats = scan.load_reduced_flats(return_info=True)
        position_in_list = min(position_in_list, len(list(reduced_flats.keys())) - 1)
        my_flat = reduced_flats[list(reduced_flats.keys())[position_in_list]]
        my_current = metadata_flats.machine_current[position_in_list]
        my_count_time = metadata_flats.count_time[position_in_list]
    elif args.neutral_flat:
        my_flat = np.ones(args.img_shape, "f")
        my_current = 1.0
        my_count_time = 1.0
    elif args.flats_from_before_after:
        scan = NXtomoScan(args.scan_before, args.entry_name)
        reduced_flats_b, metadata_flats_b = scan.load_reduced_flats(return_info=True)
        reduced_darks_b, _ = scan.load_reduced_darks(return_info=True)
        scan = NXtomoScan(args.scan_after, args.entry_name)
        reduced_flats_e, metadata_flats_e = scan.load_reduced_flats(return_info=True)
        reduced_darks_e, _ = scan.load_reduced_darks(return_info=True)

        flat_b = reduced_flats_b[list(reduced_flats_b.keys())[0]]
        flat_e = reduced_flats_e[list(reduced_flats_e.keys())[0]]
        dark_b = reduced_darks_b[list(reduced_darks_b.keys())[0]]
        dark_e = reduced_darks_b[list(reduced_darks_e.keys())[0]]

        current_b = metadata_flats_b.machine_current[0]
        current_e = metadata_flats_e.machine_current[0]

        if args.used_current is None:
            args.used_current = current_b

        factor = (i_stage) / (args.total_nstages)
        my_flat = args.used_dark + (
            (flat_b - dark_b) * (args.used_current / current_b) * (1 - factor)
            + (flat_e - dark_e) * (args.used_current / current_e) * factor
        )
        my_current = args.used_current
        my_count_time = args.used_count_time
    else:
        raise ValueError(
            "one of the option must be activated in[neutral_flat, flats_from_reduced, flats_from_before_after]"
        )

    flats_dictionary[actual_pos] = my_flat
    flats_infos_dictionary["machine_current"].append(my_current)
    flats_infos_dictionary["count_time"].append(my_count_time)


def add_dark_to_dict(
    args, darks_dictionary, darks_infos_dictionary, i_stage, actual_pos
):
    if args.flats_from_reduced:
        filename = args.file_list[i_stage - args.first_stage]
    elif args.flats_from_before_after:
        filename = args.scan_before
    else:
        assert args.neutral_flat
        filename = None

    if filename is not None:
        scan = NXtomoScan(filename, args.entry_name)
        reduced_darks, metadata_darks = scan.load_reduced_darks(return_info=True)
        my_dark = reduced_darks[list(reduced_darks.keys())[0]]
        my_count_time = metadata_darks.count_time[0]
    else:
        my_dark = np.zeros(args.img_shape, "f")
        my_count_time = 1.0

    darks_dictionary[actual_pos] = my_dark
    darks_infos_dictionary["count_time"].append(my_count_time)
    args.used_count_time = my_count_time

    args.used_dark = darks_dictionary[actual_pos]



# coding: utf-8

"""
Application to convert a bliss-hdf5 tomography dataset to Nexus - NXtomo (hdf5) format

.. program-output:: nxtomomill h52nx --help

For a complete tutorial you can have a look at: :ref:`Tomoh52nx`

"""

import argparse
import logging

from tqdm import tqdm

from nxtomomill import utils
from nxtomomill.converter import from_h5_to_nx
from nxtomomill.io.config.confighandler import (
    SETTABLE_PARAMETERS_UNITS,
    TomoHDF5ConfigHandler,
)

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


def _getPossibleInputParams():
    """

    :return: string with param1 (expected unit) ...
    """
    res = []
    for key, value in SETTABLE_PARAMETERS_UNITS.items():
        res.append(f"{key} ({value})")
    return ", ".join(res)


def _ask_for_selecting_detector(det_grps: list[str] | tuple[str]) -> str | None:
    res = input(
        "Several detector found. Only one detector is managed at the "
        "time. Please enter the name of the detector you want to use "
        f"or 'Cancel' to stop translation ({det_grps})"
    )
    if res == "Cancel":
        return None
    elif res in det_grps:
        return res
    else:
        # warning: this is not a debug log !!!
        print("detector name not recognized.")
        return _ask_for_selecting_detector(det_grps)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="convert data acquired as "
        "hdf5 from bliss to nexus "
        "`NXtomo` classes. For `zseries` it will create one entry per `z`"
    )
    parser.add_argument(
        "input_file", help="master file of the " "acquisition", default=None, nargs="?"
    )
    parser.add_argument(
        "output_file", help="output .nx or .h5 file", default=None, nargs="?"
    )
    parser.add_argument(
        "--file-extension",
        "--file_extension",
        default=None,
        help="extension of the output file. Valid values are "
        "" + "/".join([item.value for item in utils.FileExtension]),
    )
    parser.add_argument(
        "--single-file",
        help="merge all scan sequence to the same output file. "
        "By default create one file per sequence and "
        "group all sequence in the output file",
        dest="single_file",
        action="store_true",
        default=None,  # default is None and not False for the ConfigHandler so it can knows when set by the user or not
    )
    parser.add_argument(
        "--with-master-file",
        help="Creates a master file linking all .nx files",
        dest="no_master_file",
        action="store_false",
        default=None,
    )
    parser.add_argument(
        "--overwrite",
        help="Do not ask for user permission to overwrite output files",
        action="store_true",
        default=None,  # default is None and not False for the ConfigHandler so it can knows when set by the user or not
    )
    parser.add_argument(
        "--debug",
        help="Set logs to debug mode",
        action="store_true",
        default=None,  # default is None and not False for the ConfigHandler so it can knows when set by the user or not
    )
    parser.add_argument(
        "--entries",
        help="Specify (root) entries to be converted. By default it will try "
        "to convert all existing entries.",
        default=None,
    )
    parser.add_argument(
        "--sub-entries-to-ignore",
        "--ignore-sub-entries",
        help="Specify (none-root) sub entries to ignore.",
        default=None,
    )
    parser.add_argument(
        "--raises-error",
        help="Raise errors if some data are not met instead of providing some"
        " default values",
        action="store_true",
        default=None,  # default is None and not False for the ConfigHandler so it can knows when set by the user or not
    )
    parser.add_argument(
        "--field-of-view",
        help="Force the output to be `Half`  or `Full` acquisition. Otherwise "
        "parse raw data to find this information.",
        default=None,
    )
    parser.add_argument(
        "--no-input",
        "--no-input-for-missing-information",
        help="The user won't be ask for any inputs",
        dest="request_input",
        action="store_false",
        default=None,  # default is None and not False for the ConfigHandler so it can knows when set by the user or not
    )
    parser.add_argument(
        "--data-copy",
        "--copy-data",
        help="Force data duplication of frames. This will permit to have an "
        "'all-embed' file. Otherwise the detector/data dataset will haves "
        "links to other files.",
        action="store_true",
        dest="default_data_copy",
        default=None,  # default is None and not False for the ConfigHandler so it can knows when set by the user or not
    )
    parser.add_argument(
        "--sample_x_keys",
        "--sample-x-keys",
        default=None,
        help="x translation key in bliss HDF5 file",
    )
    parser.add_argument(
        "--sample_y_keys",
        "--sample-y-keys",
        default=None,
        help="y translation key in bliss HDF5 file",
    )
    parser.add_argument(
        "--translation_z_keys",
        "--translation-z-keys",
        default=None,
        help="z translation key in bliss HDF5 file",
    )
    parser.add_argument(
        "--sample-detector-distance-keys",
        "--sample-detector-distance-paths",
        "--distance-paths",
        default=None,
        help="sample detector distance paths",
    )
    parser.add_argument(
        "--valid_camera_names",
        "--valid-camera-names",
        default=None,
        help="Valid NXDetector dataset name to be considered. Otherwise will"
        "try to deduce them from NX_class attibute (value should be"
        "NXdetector) or from instrument group child structure.",
    )
    parser.add_argument(
        "--rotation_angle_keys",
        "--rotation-angle-keys",
        "--rot-angle-keys",
        "--rot_angle_keys",
        default=None,
        help="Valid dataset name for rotation angle",
    )
    parser.add_argument(
        "--exposure_time_keys",
        "--exposure-time-keys",
        "--acq_expo_time_keys",
        "--acq-expo-time-keys",
        default=None,
        help="Valid dataset name for acquisition exposure time",
    )
    parser.add_argument(
        "--sample-x-pixel-size_keys",
        "--sample_x_pixel_size_keys",
        default=None,
        help="X pixel size key to read",
    )
    parser.add_argument(
        "--sample-y-pixel-size_keys",
        "--sample_y_pixel_size_keys",
        default=None,
        help="Y pixel size key to read",
    )

    # scan titles
    parser.add_argument(
        "--init_titles",
        "--init-titles",
        default=None,
        help="Titles corresponding to init scans",
    )
    parser.add_argument(
        "--zseries-init-titles",
        "--zseries_init_titles",
        "--z_series_init_titles",
        "--z-series_init_titles",
        "--init_zserie_titles",
        "--init-zserie-titles",
        default=None,
        help="Titles corresponding to zserie init scans",
    )
    parser.add_argument(
        "--multitomo-init-titles",
        "--multitomo_init_titles",
        "--multi-tomo-init-titles",
        "--multi_tomo_init_titles",
        "--init-multi-tomo-titles",
        "--init-pcotomo-titles",
        default=None,
        help="Titles corresponding to multi-tomo init scans",
    )
    parser.add_argument(
        "--back-and-forth-init-titles",
        "--back_and_forth_init_titles",
        "--init-back-and-forth-titles",
        "--init_back_and_forth_titles",
        default=None,
        help="Titles corresponding to back-and-forth init scans",
    )
    parser.add_argument(
        "--dark-titles",
        "--dark_titles",
        default=None,
        help="Titles corresponding to dark scans",
    )
    parser.add_argument(
        "--flat-titles",
        "--flat_titles",
        "--ref_titles",
        "--ref-titles",
        default=None,
        help="Titles corresponding to ref scans",
        dest="flat_titles",
    )
    parser.add_argument(
        "--projection_titles",
        "--proj_titles",
        "--proj-titles",
        default=None,
        help="Titles corresponding to projection scans",
    )
    parser.add_argument(
        "--alignment-titles",
        "--alignment_titles",
        "--align_titles",
        "--align-titles",
        default=None,
        help="Titles corresponding to alignment scans",
    )
    parser.add_argument(
        "--set-params",
        default=None,
        nargs="*",
        help="Allow manual definition of some parameters. "
        "Valid parameters (and expected input unit) "
        f"are: {_getPossibleInputParams()}. Should be added at the end of the command line because "
        "will try to cover all text set after this option.",
    )

    # config file
    parser.add_argument(
        "--config-file",
        "--config",
        "--configuration",
        "--configuration-file",
        default=None,
        help="file containing the full configuration to convert from h5 "
        "bliss to nexus",
    )

    options = parser.parse_args(argv[1:])
    if options.request_input:
        callback_det_sel = _ask_for_selecting_detector
    else:
        callback_det_sel = None
    try:
        configuration_handler = TomoHDF5ConfigHandler(options)
    except Exception:
        _logger.error("Fail to initiate the configuration.", exc_info=True)
        return
    else:
        for title in configuration_handler.configuration.init_titles:
            assert title != ""
        logging.getLogger("nxtomomill").setLevel(
            configuration_handler.configuration.log_level
        )
        from_h5_to_nx(
            configuration=configuration_handler.configuration,
            progress=tqdm(desc="h52nx", delay=1, bar_format="{l_bar}{bar}"),
            input_callback=None,
            detector_sel_callback=callback_det_sel,
        )



"""application to copy NXtomo(s) from one file to another"""

import argparse
import os
from nxtomo.application.nxtomo import copy_nxtomo_file as copy_nxtomo
from nxtomomill.models.utils import convert_str_to_tuple
from nxtomomill.utils._os import abspath


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="copy one or several NXtomo to another location"
    )
    parser.add_argument(
        "nexus_file", help="file path to the nexus file containing NXtomo", nargs="?"
    )
    parser.add_argument("output_file", help="output nexus file", nargs="?")
    parser.add_argument(
        "--entry",
        "--entries",
        default=None,
        help="NXtomo path(s) to be copied. If none provided then all NXtomo entries will be copied",
        dest="entries",
    )
    parser.add_argument(
        "--overwrite",
        help="Do not ask for user permission to overwrite output files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug",
        help="Set logs to debug mode",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--remove-vds",
        help="Remove any Virtual dataset to the resulting NXtomo (duplicate detector data - warning: all data will be load in memory before dumping it)",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(argv[1:])
    copy_nxtomo(
        input_file=options.nexus_file,
        output_file=get_output_file(options.output_file, options.nexus_file),
        entries=(
            convert_str_to_tuple(options.entries)
            if options.entries is not None
            else None
        ),
        overwrite=options.overwrite,
        vds_resolution="remove" if options.remove_vds else "update",
    )


def get_output_file(output_file_or_folder, input_file) -> str:
    """compute output file for copying NXtomo from input file or folder"""

    def get_output_file_from_folder():
        return os.path.join(
            output_file_or_folder, os.path.basename(abspath(input_file))
        )

    if os.path.isdir(output_file_or_folder):
        output_file = get_output_file_from_folder()
    elif (
        os.path.isfile(output_file_or_folder)
        or os.path.splitext(output_file_or_folder)[-1] != ""
    ):
        output_file = output_file_or_folder
    else:
        output_file = get_output_file_from_folder()
    return abspath(output_file)



import os
from nxtomo.application.nxtomo import NXtomo
from nxtomomill.app.nxcopy import get_output_file
from nxtomomill.app.nxcopy import main


def test_nxcopy_get_output_file():
    """dummy test for 'get_output_file'"""
    assert get_output_file("output.nx", "input.nx") == os.path.abspath("output.nx")
    assert (
        get_output_file("/file1/to/", "/file2/to/input/input.nx")
        == "/file1/to/input.nx"
    )


def test_copy(tmp_path):
    """test 'copy' application"""
    input_folder = tmp_path / "input"
    input_folder.mkdir()
    input_nx_tomo_file = os.path.join(input_folder, "nexus.nx")

    output_folder = tmp_path / "output"
    output_folder.mkdir()
    output_folder = str(output_folder)

    nx_tomo = NXtomo()
    nx_tomo.save(input_nx_tomo_file, "/entry0000")
    nx_tomo.save(input_nx_tomo_file, "/entry0002")
    nx_tomo.save(input_nx_tomo_file, "/entry0004")

    main(["nx-copy", input_nx_tomo_file, output_folder, "--entry", "entry0002"])
    output_file = os.path.join(output_folder, "nexus.nx")
    assert os.path.exists(output_file)
    assert len(NXtomo.get_valid_entries(output_file)) == 1
    main(["nx-copy", input_nx_tomo_file, output_folder, "--overwrite"])
    assert len(NXtomo.get_valid_entries(output_file)) == 3
    main(["nx-copy", input_nx_tomo_file, output_folder, "--overwrite", "--remove-vds"])
    assert len(NXtomo.get_valid_entries(output_file)) == 3



from __future__ import annotations

from nxtomomill.models.utils import convert_str_to_tuple, convert_str_to_bool
from nxtomomill.settings import Tomo
from nxtomomill.models.utils import PathType
from nxtomomill.models.utils import filter_str_def

from ..base.general_section import GeneralSection as _GeneralSectionBase

from pydantic import Field, field_validator, ConfigDict, field_serializer


class GeneralSection(_GeneralSectionBase):
    model_config: ConfigDict = ConfigDict(
        validate_assignment=True,
        validate_by_name=True,
    )

    input_folder: str | None = Field(
        default=None,
        description="Path to the folder containing .edf files (mandatory)",
    )
    delete_edf_source_files: bool = Field(
        default=False,
        description="Remove EDF source files after successful conversion. This operation is conditional and will only execute if the 'duplicate_data' flag is set to True. By default, 'duplicate_data' is set to True, allowing this cleanup operation to proceed",
        alias="delete_edf_source_file",
    )
    output_checks: tuple[str, ...] = Field(
        default=tuple(),
        description="Perform validation checks post-conversion to ensure data integrity and accuracy."
        "This function accepts a list of specified tests to validate the conversion process."
        "Currently supported test:\n"
        "- 'compare-output-volume': Compares the volume of the output data to ensure it matches expected values.",
    )
    dataset_basename: str | None = Field(
        default=None,
        description="Prefix for dataset files used to identify associated projection and info files."
        "If this prefix is not explicitly provided, the system will default to using the name of the input folder as the prefix."
        "This ensures that all related files can be accurately linked and processed to",
    )
    dataset_info_file: str | None = Field(
        default=None,
        description="Path to the .info file containing metadata about the dataset, such as Energy, ScanRange, and TOMO_N."
        "If this path is not provided, the system will attempt to deduce it automatically using the dataset_basename."
        "This ensures that necessary metadata is accessible for further processing and analysis.",
    )
    title: str | None = Field(default=None, description="NXtomo title")
    patterns_to_ignores: tuple[str, ...] = Field(
        default=Tomo.EDF.TO_IGNORE,
        description="Specifies file patterns that determine which files should be ignored during processing."
        "Files matching these patterns will be excluded from operations. "
        "This is particularly useful for filtering out unnecessary or intermediate files, such as reconstructed slice files.",
    )
    duplicate_data: bool = Field(
        default=True,
        description="Determines if we can create files with external links or not."
        "If False then will create embed all the data into a single file avoiding external link to other file. "
        "If True then the detector data will point to original tif files. In this case you must be cautious to keep relative paths valid. Warning: to read external dataset you nust be at the hdf5 file working directory. See external link resolution details: https://support.hdfgroup.org/documentation/hdf5/latest/group___h5_l.html#title5",
    )
    external_link_type: PathType = Field(
        default=PathType.RELATIVE,
        description="Specifies the type of file path to use for linking to original files when 'duplicate_data' is set to False."
        "When 'duplicate_data' is False, you can choose how file paths are referenced:"
        "\n- 'relative': Uses relative paths for linking to the original files."
        "\n- 'absolute': Uses absolute paths for linking to the original files.",
        alias="external_link_path",
    )

    @field_validator(
        "input_folder",
        "dataset_basename",
        "dataset_info_file",
        "title",
        mode="plain",
    )
    @classmethod
    def cast_to_str(cls, value: str | None) -> tuple | None:
        return filter_str_def(value)

    @field_validator(
        "patterns_to_ignores",
        "output_checks",
        mode="plain",
    )
    @classmethod
    def cast_to_tuple(cls, value: tuple[str, ...] | str | None) -> tuple[str]:
        if isinstance(value, tuple):
            return value
        value = filter_str_def(value)
        return convert_str_to_tuple(value) or ()

    @field_validator(
        "external_link_type",
    )
    @classmethod
    def cast_to_path_type(cls, value: str | PathType) -> PathType:
        if isinstance(value, str):
            value = filter_str_def(value).lower()
        return PathType(value)

    @field_serializer(
        "external_link_type",
    )
    @classmethod
    def serialize_external_link_type(cls, value: PathType):
        return value.value

    @field_validator(
        "delete_edf_source_files",
        mode="plain",
    )
    @classmethod
    def cast_to_bool(cls, value: bool | str) -> bool:
        if isinstance(value, str):
            value = filter_str_def(value)
        return convert_str_to_bool(value)



from __future__ import annotations

import os
import shutil
import pytest
import h5py
import numpy

from nxtomo import NXtomo
from nxtomo.utils.transformation import DetXFlipTransformation, DetYFlipTransformation
from nxtomomill.models.h52nx import H52nxModel
from nxtomomill.converter.hdf5.hdf5converter import from_h5_to_nx
from nxtomomill.tests.datasets import GitlabDataset


@pytest.fixture()
def dataset_id16a(tmp_path):
    source_file = GitlabDataset.get_dataset(
        "h5_datasets/holotomo/id16a/Atomium_S2_holo4_HT_010nm_0001.h5"
    )
    return shutil.copyfile(
        source_file,
        os.path.join(tmp_path, "Atomium_S2_holo1_HT_010nm_500proj_0003.h5"),
    )


@pytest.mark.parametrize("soft_flip", ((True, False), (False, True)))
@pytest.mark.parametrize("mechanical_flip", ((False, False), (True, True)))
def test_frame_flips(tmp_path, dataset_id16a, soft_flip, mechanical_flip):
    # edit
    with h5py.File(dataset_id16a, mode="a") as h5f:
        tomo_config = h5f["1.1/technique/detector/balor"]
        del tomo_config["flipping"]
        tomo_config["flipping"] = numpy.array(soft_flip)

    output_file = os.path.join(tmp_path, "my_nx.nx")
    assert not os.path.exists(output_file)

    # convert
    model = H52nxModel(
        input_file=dataset_id16a,
        output_file=output_file,
        mechanical_lr_flip=mechanical_flip[0],
        mechanical_ud_flip=mechanical_flip[1],
        single_file=True,
    )

    # check transformations matrix
    from_h5_to_nx(configuration=model)

    assert os.path.exists(output_file)

    nxtomo = NXtomo().load(file_path=output_file, data_path="entry0000")
    transformations = nxtomo.instrument.detector.transformations.transformations

    final_lr_flip = soft_flip[0] ^ mechanical_flip[0]
    final_ud_flip = soft_flip[1] ^ mechanical_flip[1]

    lr_flip_mat = DetXFlipTransformation(flip=final_ud_flip)
    ud_flip_mat = DetYFlipTransformation(flip=final_lr_flip)

    assert lr_flip_mat in transformations
    assert ud_flip_mat in transformations



"""
Application to split a file containing several NXtomo entries into several files containing each a single NXtomo

.. program-output:: split-nxfile --help

"""

import os
import argparse
import logging
import string

from nxtomo.application.nxtomo import NXtomo
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from silx.io.utils import open as open_hdf5
from nxtomomill.utils._os import abspath

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="split a file containing several NXtomo (at root level) into multiple files containing each a single NXtomo"
    )
    parser.add_argument(
        "input_file",
        help="File containing NXtomo to be split into several files",
        nargs="?",
    )
    parser.add_argument(
        "output_file_pattern",
        help="output file pattern. Must contain `{entry_name}` or `{index}` pattern to make sure it is unique",
        default="{input_file_name}_{entry_name}.nx",
        nargs="?",
    )

    parser.add_argument(
        "--overwrite",
        help="Do not ask for user permission to overwrite output files",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--duplicate-data",
        help="Make all NXtomo free of any external link. As a result this will duplicate data",
        nargs="?",
    )

    options = parser.parse_args(argv[1:])
    output_file_pattern = options.output_file_pattern
    if "{input_file_name}" in output_file_pattern:
        output_file_pattern = output_file_pattern.format(
            {
                "output_file_pattern": os.path.splitext(options.input_file),
            }
        )

    split(
        input_file=options.input_file,
        output_file_pattern=output_file_pattern,
        overwrite=options.overwrite,
    )


def split(
    input_file: str,
    output_file_pattern: str,
    overwrite: bool = False,
    duplicate_data: bool = False,
) -> tuple:
    """
    :param input_file: path to the file to be splitted
    :param output_file_pattern: pattern of the file to create. Must contain either `{entry_name}` or `{index}`
    :param overwrite: can we overwrite output file if alread exists

    :return: tuple of identifier of all the NXtomo generated
    """

    def get_output_file(index: int, entry_name: str, pattern: str) -> str:
        """
        treat 'pattern' to create the expected output file
        """
        keywords = {
            "entry_name": entry_name,
            "index": index,
        }

        # filter necessary keywords
        def get_necessary_keywords():
            formatter = string.Formatter()
            return [field for _, field, _, _ in formatter.parse(pattern) if field]

        requested_keywords = get_necessary_keywords()

        def keyword_needed(pair):
            keyword, _ = pair
            return keyword in requested_keywords

        keywords = dict(filter(keyword_needed, keywords.items()))
        if len(keywords) == 0:
            raise ValueError(
                "pattern should at least contains keywords '{index}' or '{entry_name}' to be format. Else unable to create a unique file per NXtomo"
            )
        return abspath(pattern.format(**keywords))

    if duplicate_data:
        detector_data_as = "as_numpy_array"
    else:
        detector_data_as = "as_data_url"
    result = []
    with open_hdf5(input_file) as h5f:
        for i_entry, entry in enumerate(h5f.keys()):
            try:
                nx_tomo = NXtomo().load(
                    input_file, entry, detector_data_as=detector_data_as
                )
            except Exception as e:
                _logger.error(
                    f"Fail to treat entry {entry}. Error is {e}. Is this a valid Nxtomo ?"
                )
            else:
                output_file = get_output_file(
                    index=i_entry, entry_name=entry, pattern=output_file_pattern
                )
                dirname = os.path.dirname(output_file)
                if dirname and not os.path.exists(dirname):
                    os.makedirs(dirname)
                nx_tomo.save(
                    file_path=output_file, data_path=entry, overwrite=overwrite
                )
                result.append(NXtomoScan(output_file, entry))

    return tuple(result)



from nxtomomill.converter.hdf5.utils import (
    get_default_output_file,
    PROCESSED_DATA_DIR_NAME,
    RAW_DATA_DIR_NAME,
)


def test_get_default_output_file():
    """
    test the get_default_output_file function
    """
    # 1. simple test
    assert get_default_output_file("/my_tmp/path/file.h5") == "/my_tmp/path/file.nx"
    assert (
        get_default_output_file(f"/my_tmp/{RAW_DATA_DIR_NAME}/file.h5")
        == f"/my_tmp/{PROCESSED_DATA_DIR_NAME}/file.nx"
    )

    assert (
        get_default_output_file(f"/my_tmp/path/{RAW_DATA_DIR_NAME}/toto/file.h5")
        == f"/my_tmp/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )
    # note: _RAW_DATA_DIR_NAME part of the path but not a folder
    assert (
        get_default_output_file(f"/my_tmp/path_{RAW_DATA_DIR_NAME}/toto/file.h5")
        == f"/my_tmp/path_{RAW_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2. advance test
    # 2.1 use case: '_RAW_DATA_DIR_NAME' is present twice in the path -> replace the deeper one
    assert (
        get_default_output_file(
            f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2.2 use case: contains both '_RAW_DATA_DIR_NAME' and '_PROCESSED_DATA_DIR_NAME' in the path
    assert (
        get_default_output_file(
            f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )

    assert (
        get_default_output_file(
            f"/my_tmp/{PROCESSED_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/my_tmp/{PROCESSED_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2.3 use case: expected output file is the input file. Make sure append '_nxtomo'
    assert (
        get_default_output_file("/my_tmp/path/file.nx") == "/my_tmp/path/file_nxtomo.nx"
    )



# coding: utf-8

"""
module to define a standard tomography acquisition (made by bliss)
"""

from __future__ import annotations

import pint
from datetime import datetime

import h5py
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset

from nxtomo.utils.transformation import DetXFlipTransformation, DetYFlipTransformation
from nxtomo.nxobject.nxsource import SourceType
from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.nxobject.utils.concatenate import (
    concatenate_pint_quantities as _concatenate_pint_quantities,
)

from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.utils.utils import str_datetime_to_numpy_datetime64
from nxtomomill.utils.io import deprecated
from nxtomomill.utils.hdf5 import get_dataset_unit
from nxtomomill.utils._os import abspath

from .baseacquisition import BaseAcquisition, EntryReader
from .utils import (
    deduce_machine_current,
    get_bliss_scan_type,
    get_nx_detectors,
    guess_nx_detector,
)

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
import fnmatch
import logging
import os

import numpy

from nxtomomill.converter.hdf5.acquisition.blisstomoconfig import (
    TomoConfig as BlissTomoConfig,
)
from nxtomomill.io.config import TomoHDF5Config
from nxtomo.application.nxtomo import NXtomo

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)


__all__ = ["StandardAcquisition"]


class StandardAcquisition(BaseAcquisition):
    """
    Class to collect information from a bliss - hdf scan (see https://bliss.gitlab-pages.esrf.fr/fscan).
    Once all data is collected a set of NXtomo will be created.
    Then NXtomo instances will be saved to disk.

    :param root_url: url of the acquisition. Can be None if
                             this is the initialization entry
    :param configuration: configuration to use to collect raw data and generate outputs
    :param detector_sel_callback: possible callback to retrieve missing information
    """

    def __init__(
        self,
        root_url: DataUrl | None,
        configuration: TomoHDF5Config,
        detector_sel_callback,
        start_index,
        parent=None,
    ):
        super().__init__(
            root_url=root_url,
            configuration=configuration,
            detector_sel_callback=detector_sel_callback,
            start_index=start_index,
        )
        self._parent = parent
        # possible parent. Like for z series
        self._nx_tomos = [NXtomo()]
        self._image_key_control = None
        self._rotation_angle: pint.Quantity | None = None
        self._sample_x: pint.Quantity | None = None
        self._sample_y: pint.Quantity | None = None
        self._translation_y: pint.Quantity | None = None
        self._translation_z: pint.Quantity | None = None
        self._lr_flipped: bool | None = None
        self._ud_flipped: bool | None = None

        self._unique_detector_names = list()
        # register names

        self._virtual_sources = None
        self._acq_expo_time: pint.Quantity | None = None
        self._copied_dataset = {}
        "register dataset copied. Key if the original location as" "DataUrl.path. Value is the DataUrl it has been moved to"
        self._known_machine_current_am: None | dict = None
        # store all registered machine current
        self._frames_timestamp = None
        # try to deduce time stamp of each frame

    def parent_root_url(self) -> DataUrl | None:
        if self._parent is not None:
            return self._parent.root_url
        else:
            return None

    def get_expected_nx_tomo(self):
        return 1

    @property
    def image_key_control(self):
        return self._image_key_control

    @property
    def rotation_angle(self):
        return self._rotation_angle

    @property
    def translation_y(self):
        return self._translation_y

    @property
    def translation_z(self):
        return self._translation_z

    @property
    def lr_flipped(self):
        return self._lr_flipped

    @deprecated(
        replacement="lr_flipped",
        reason="renamed",
        since_version="2.0",
    )
    @property
    def x_flipped(self):
        return self.lr_flipped

    @property
    def ud_flipped(self):
        return self._ud_flipped

    @deprecated(
        replacement="ud_flipped",
        reason="renamed",
        since_version="2.0",
    )
    @property
    def y_flipped(self):
        return self.ud_flipped

    @property
    def n_frames(self):
        return self._n_frames

    @property
    def n_frames_actual_bliss_scan(self):
        return self._n_frames_actual_bliss_scan

    @property
    def dim_1(self):
        return self._dim_1

    @property
    def dim_2(self):
        return self._dim_2

    @property
    def data_type(self):
        return self._data_type

    @property
    def expo_time(self):
        return self._acq_expo_time

    @property
    def known_machine_current(self) -> dict | None:
        """
        Return the dict of all known machine currents. Key is the time stamp, value is the machine current
        """
        return self._known_machine_current_am

    @property
    def sample_x(self):
        """Return the '_sample_x' attribute. In **ESRF coordinate**"""
        return self._sample_x

    @property
    def sample_y(self):
        """Return the '_sample_y' attribute. In **ESRF coordinate**"""
        return self._sample_y

    def register_step(
        self, url: DataUrl, entry_type: AcquisitionStep | None = None, copy_frames=False
    ) -> None:
        """

        :param url: entry to be registered and contained in the
                                 acquisition
        :param entry_type: type of the entry if know. Overwise will be
                           'evaluated'
        """
        if entry_type is None:
            entry_type = get_bliss_scan_type(url=url, configuration=self.configuration)
        assert (
            entry_type is not AcquisitionStep.INITIALIZATION
        ), "Initialization are root node of a new sequence and not a scan of a sequence"

        if entry_type is None:
            _logger.warning(f"{url} not recognized, skip it")
        else:
            self._registered_entries[url.path()] = entry_type
            self._copy_frames[url.path()] = copy_frames
            self._entries_o_path[url.path()] = url.data_path()
            # path from the original file. Haven't found another way to get it ?!

    def _get_valid_camera_names(self, instrument_grp: h5py.Group):
        # 1: try to get detector from nx property
        detectors = get_nx_detectors(instrument_grp)
        detectors = [grp.name.split("/")[-1] for grp in detectors]

        def filter_detectors(det_grps):
            if len(det_grps) > 0:
                _logger.info(f"{len(det_grps)} detector found from NX_class attribute")
                if len(det_grps) > 1:
                    # if an option: pick the first one once orderered
                    # else ask user
                    if self._detector_sel_callback is None:
                        sel_det = det_grps[0]
                        _logger.warning(
                            f"several detector found. Only one is managed for now. Will pick {sel_det}"
                        )
                    else:
                        sel_det = self._detector_sel_callback(det_grps)
                        if sel_det is None:
                            _logger.warning("no detector given, avoid conversion")
                    det_grps = (sel_det,)
                return det_grps
            return None

        detectors = filter_detectors(det_grps=detectors)
        if detectors is not None:
            return detectors

        # 2: get nx detector from shape...
        detectors = guess_nx_detector(instrument_grp)
        detectors = [grp.name.split("/")[-1] for grp in detectors]
        return filter_detectors(det_grps=detectors)

    @staticmethod
    def concatenate_pint_quantities(quantities: tuple[pint.Quantity | None]):
        """
        concatenation dedicated to acquisition. quantities items can be None or a quantity
        """
        return _concatenate_pint_quantities(
            tuple(
                filter(
                    lambda a: a is not None and len(a) > 0,
                    quantities,
                )
            )
        )

    def __get_data_from_camera(
        self,
        data_dataset: h5py.Dataset,
        data_name,
        frame_type,
        entry,
        entry_path,
        camera_dataset_url,
    ):
        if data_dataset.ndim == 2:
            shape = (1, data_dataset.shape[0], data_dataset.shape[1])
        elif data_dataset.ndim != 3:
            err = f"dataset {data_name} is expected to be 3D when {data_dataset.ndim}D found."
            if data_dataset.ndim == 1:
                err = "\n".join(
                    [
                        err,
                        "This might be a bliss-EDF dataset. Those are not handled by nxtomomill",
                    ]
                )
            _logger.error(err)
            return 0
        else:
            shape = data_dataset.shape

        n_frame = shape[0]
        self._n_frames += n_frame
        self._n_frames_actual_bliss_scan = n_frame
        if self.dim_1 is None:
            self._dim_2 = shape[1]
            self._dim_1 = shape[2]
        else:
            if self._dim_1 != shape[2] or self._dim_2 != shape[1]:
                raise ValueError("Inconsistency in detector shapes")
        if self._data_type is None:
            self._data_type = data_dataset.dtype
        elif self._data_type != data_dataset.dtype:
            raise ValueError("detector frames have incoherent " "data types")

        # update image_key and image_key_control
        # Note: for now there is no image_key on the master file
        # should be added later.
        image_key_control = frame_type.to_image_key_control()
        self._image_key_control.extend([image_key_control.value] * n_frame)

        data_dataset_path = data_dataset.name.replace(entry.name, entry_path, 1)
        # replace data_dataset name by the original entry_path.
        # this is a workaround to use the dataset path on the
        # "treated file". Because .name if the name on the 'target'
        # file of the virtual dataset
        v_source = h5py.VirtualSource(
            camera_dataset_url.file_path(),
            data_dataset_path,
            data_dataset.shape,
            dtype=self._data_type,
        )
        self._virtual_sources.append(v_source)
        self._virtual_sources_len.append(n_frame)
        return n_frame

    def _treate_valid_camera(
        self, detector_node, entry, frame_type, input_file_path, entry_path, entry_url
    ) -> bool:
        """
        treate a dataset considered as a 'camera' dataset.
        """
        if "data_cast" in detector_node:
            _logger.warning(
                f"!!! looks like this data has been cast. Take cast data for {detector_node}!!!"
            )
            data_dataset = detector_node["data_cast"]
            data_name = "/".join((detector_node.name, "data_cast"))
        elif "data" in detector_node:
            data_dataset = detector_node["data"]
            data_name = "/".join((detector_node.name, "data"))
        else:
            raise KeyError(f"Unable to find camera dataset in {detector_node.name}")

        camera_dataset_url = DataUrl(
            file_path=entry_url.file_path(), data_path=data_name, scheme="silx"
        )

        n_frame = self.__get_data_from_camera(
            data_dataset,
            data_name=data_name,
            frame_type=frame_type,
            entry=entry,
            entry_path=entry_path,
            camera_dataset_url=camera_dataset_url,
        )
        # save information if this url must be embed / copy or not. Will be used later at nxtomo side
        self._copy_frames[camera_dataset_url.path()] = self._copy_frames[
            entry_url.path()
        ]

        soft_lr_flip, soft_ud_flip = self._get_soft_flip_frame()
        # handle None flip values
        soft_lr_flip = soft_lr_flip or False
        soft_ud_flip = soft_ud_flip or False
        # get the final flip value (soft + mechanical)
        lr_flipped = soft_lr_flip ^ self.configuration.mechanical_lr_flip
        ud_flipped = soft_ud_flip ^ self.configuration.mechanical_ud_flip

        if self._lr_flipped is None and self._ud_flipped is None:
            # if this is the first scan then define frame flips
            self._lr_flipped, self._ud_flipped = bool(lr_flipped), bool(ud_flipped)
        elif lr_flipped != self._lr_flipped or ud_flipped != self._ud_flipped:
            # else check it is consistent with already existing flips
            raise ValueError(
                f"Found different detector flips inside the same sequence on {entry}. Unable to handle it."
            )

        # store rotation
        rots = self._get_rotation_angle(root_node=entry, n_frame=n_frame)
        # handle rotation_is_clockwise. It None or False consider it is counter-clockwise
        # NXtomo coordinate system uses MCstas that uses counter-clockwise rotation direction.
        # https://manual.nexusformat.org/design.html#the-nexus-coordinate-system
        if self.configuration.rotation_is_clockwise is True:
            rots *= -1.0
        self._rotation_angle = StandardAcquisition.concatenate_pint_quantities(
            (
                self._rotation_angle,
                rots,
            ),
        )

        self._sample_x = StandardAcquisition.concatenate_pint_quantities(
            (
                self._sample_x,
                self._get_sample_x(root_node=entry, n_frame=n_frame),
            ),
        )

        self._sample_y = StandardAcquisition.concatenate_pint_quantities(
            (
                self._sample_y,
                self._get_sample_y(root_node=entry, n_frame=n_frame),
            ),
        )

        self._translation_y = StandardAcquisition.concatenate_pint_quantities(
            (
                self._translation_y,
                self._get_translation_y(root_node=entry, n_frame=n_frame),
            ),
        )

        self._translation_z = StandardAcquisition.concatenate_pint_quantities(
            (
                self._translation_z,
                self._get_translation_z(root_node=entry, n_frame=n_frame),
            ),
        )

        # store acquisition time
        self._acq_expo_time = StandardAcquisition.concatenate_pint_quantities(
            (
                self._acq_expo_time,
                self._get_expo_time(
                    root_node=entry,
                    detector_node=detector_node,
                    n_frame=n_frame,
                ),
            )
        )

        self._current_scan_n_frame = n_frame

    def camera_is_valid(self, det_name):
        assert isinstance(det_name, str)
        if len(self.configuration.valid_camera_names) == 0:
            return True
        for vcm in self.configuration.valid_camera_names:
            if fnmatch.fnmatch(det_name, vcm):
                return True
        return False

    def _preprocess_registered_entry(self, entry_url, type_):
        with EntryReader(entry_url) as entry:
            entry_path = self._entries_o_path[entry_url.path()]
            input_file_path = entry_url.file_path()
            input_file_path = abspath(os.path.relpath(input_file_path, os.getcwd()))
            input_file_path = abspath(input_file_path)
            if type_ is AcquisitionStep.INITIALIZATION:
                raise RuntimeError(
                    "no initialization should be registered."
                    "There should be only one per acquisition."
                )

            if "instrument" not in entry:
                _logger.error(
                    f"no instrument group found in {entry.name}, unable to retrieve frames"
                )
                return

            instrument_grp = entry["instrument"]

            # if we don't get a valid camera (not provided by the user or not found on the bliss tomo metadata)
            if len(self.configuration.valid_camera_names) == 0:
                # if we need to guess detector name(s)
                # ignore in case we read information from bliss config
                det_grps = self._get_valid_camera_names(instrument_grp)
                # update valid camera names
                self.configuration.valid_camera_names = det_grps

            has_frames = False
            for key, _ in instrument_grp.items():
                if (
                    "NX_class" in instrument_grp[key].attrs
                    and instrument_grp[key].attrs["NX_class"] == "NXdetector"
                ):
                    _logger.debug(f"Found one detector at {key} for {entry.name}.")

                    if not self.camera_is_valid(key):
                        _logger.debug(f"ignore {key}, not a `valid` camera name")
                        continue
                    else:
                        detector_node = instrument_grp[key]
                        if key not in self._unique_detector_names:
                            self._unique_detector_names.append(key)
                        try:
                            self._treate_valid_camera(
                                detector_node,
                                entry=entry,
                                frame_type=type_,
                                input_file_path=input_file_path,
                                entry_path=entry_path,
                                entry_url=entry_url,
                            )
                        except KeyError as e:
                            if self.configuration.raises_error:
                                raise e
                            has_frames = False
                        else:
                            has_frames = True
            # try to get some other metadata

            # handle frame time stamp
            start_time = self._get_start_time(entry)
            if start_time is not None:
                start_time = datetime.fromisoformat(start_time)
            end_time = self._get_end_time(entry)
            if end_time is not None:
                end_time = datetime.fromisoformat(end_time)
            if has_frames:
                self._register_frame_timestamp(entry, start_time, end_time)

            # handle machine current. Can retrieve some current even on bliss scan entry not containing directly frames
            self._register_machine_current(entry, start_time, end_time)

    def _register_machine_current(self, entry: h5py.Group, start_time, end_time):
        """Update machine electric current for provided entry (bliss scan"""
        machine_currents: pint.Quantity | None = self._get_machine_current(
            root_node=entry
        )
        # electric current will be saved as Ampere
        if machine_currents is not None and len(machine_currents) > 0:
            machine_currents = machine_currents.to(_ureg.ampere).magnitude
            new_know_machine_currents_am = {}
            if start_time is None or end_time is None:
                if start_time != end_time:
                    _logger.warning(
                        f"Unable to find {'start_time' if start_time is None else 'end_time'}. Will pick the first available machine_current for the frame"
                    )
                    t_time = start_time or end_time
                    # if at least one can find out
                    new_know_machine_currents_am[
                        str_datetime_to_numpy_datetime64(t_time)
                    ] = machine_currents[0]
                else:
                    _logger.error(
                        "Unable to find start_time and end_time. Will not register any machine current"
                    )
            elif len(machine_currents) == 1:
                # if we have only one value, consider the machine current is constant during this time
                # might be improved later if we can know if current is determine at the
                # beginning or the end. But should have no impact
                # as the time slot is short
                new_know_machine_currents_am[
                    str_datetime_to_numpy_datetime64(start_time)
                ] = machine_currents[0]
            else:
                # linspace from datetime within ms precision.
                # see https://stackoverflow.com/questions/37964100/creating-numpy-linspace-out-of-datetime
                timestamps = numpy.linspace(
                    start=str_datetime_to_numpy_datetime64(start_time).astype(
                        numpy.float128
                    ),
                    stop=str_datetime_to_numpy_datetime64(end_time).astype(
                        numpy.float128
                    ),
                    num=len(machine_currents),
                    endpoint=True,
                    dtype="<M8[ms]",
                )
                for timestamp, machine_current_item in zip(
                    timestamps, machine_currents
                ):
                    new_know_machine_currents_am[timestamp.astype(numpy.datetime64)] = (
                        machine_current_item
                    )
            # filter nan values
            filtered_new_know_machine_currents = {
                key: value
                for key, value in new_know_machine_currents_am.items()
                if not numpy.isnan(value)
            }
            if len(filtered_new_know_machine_currents) != len(
                new_know_machine_currents_am
            ):
                _logger.warning(f"Found current == nan in {entry}")
            # update known_machine_current
            self._known_machine_current_am.update(filtered_new_know_machine_currents)

    def _register_frame_timestamp(self, entry: h5py.Group, start_time, end_time):
        """
        update frame time stamp for the provided entry (bliss scan)
        """
        if start_time is None or end_time is None:
            if start_time != end_time:
                t_time = str_datetime_to_numpy_datetime64(start_time or end_time)
                message = f"Unable to find start_time and / or end_time. Takes {t_time} as frame time stamp for {entry} "
                self._frames_timestamp.extend(
                    [t_time] * self._n_frames_actual_bliss_scan
                )
                _logger.warning(message)
            else:
                message = f"Unable to find start_time and end_time. Can't deduce frames time stamp for {entry}"
                _logger.error(message)
        else:
            frames_times_stamps_as_f8 = numpy.linspace(
                start=str_datetime_to_numpy_datetime64(start_time).astype(
                    numpy.float128
                ),
                stop=str_datetime_to_numpy_datetime64(end_time).astype(numpy.float128),
                num=self._n_frames_actual_bliss_scan,
                endpoint=True,
                dtype="<M8[ms]",
            )
            frames_times_stamps_as_f8 = [
                timestamp.astype("<M8[ms]") for timestamp in frames_times_stamps_as_f8
            ]
            self._frames_timestamp.extend(frames_times_stamps_as_f8)

    def _preprocess_registered_entries(self):
        """parse all frames of the different steps and retrieve data,
        image_key..."""
        self._n_frames = 0
        self._n_frames_actual_bliss_scan = 0
        # number of frame contains in X.1
        self._dim_1 = None
        self._dim_2 = None
        self._data_type = None
        self._translation_y = None
        self._translation_z = None
        self._sample_x = None
        self._sample_y = None
        self._image_key_control = []
        self._rotation_angle = None
        self._known_machine_current_am = {}
        self._frames_timestamp = []
        self._virtual_sources = []
        self._instrument_name = None
        self._virtual_sources_len = []
        self._diode = []
        self._acq_expo_time: pint.Quantity | None = None
        self._diode_unit = None
        self._copied_dataset = {}
        self._lr_flipped = None
        self._ud_flipped = None

        # if rotation motor is not defined try to deduce it from root_url/technique/scan/motor
        if self.configuration.rotation_angle_keys is None:
            rotation_motor = self._read_rotation_motor_name()
            if rotation_motor is not None:
                self.configuration.rotation_angle_keys = (rotation_motor,)
            else:
                self.configuration.rotation_angle_keys = tuple()

        # list of data virtual source for the virtual dataset
        for entry_url, type_ in self._registered_entries.items():
            url = DataUrl(path=entry_url)
            self._n_frames_actual_bliss_scan = 0
            self._preprocess_registered_entry(url, type_)

        if len(self._diode) == 0:
            self._diode = None
        if self._diode is not None:
            self._diode = numpy.asarray(self._diode)
            self._diode = self._diode / self._diode.mean()

    def _get_diode(self, root_node, n_frame) -> tuple:
        values, unit = self._get_node_values_for_frame_array(
            node=root_node["measurement"],
            n_frame=n_frame,
            keys=self.configuration.diode_keys,
            info_retrieve="diode",
            expected_unit=_ureg.volt,
        )
        return values, unit

    def _generic_path_getter(self, paths: tuple, message, level="warning", entry=None):
        """
        :param level: level can be logging.level values : "warning", "error", "info"
        :param H5group entry: user can provide directly an entry to be used as an open h5Group
        """
        if not isinstance(paths, tuple):
            raise TypeError

        url = self.parent_root_url() or self.root_url
        if url is not None:
            self._check_has_metadata(url)

        def process(h5_group):
            for path in paths:
                if h5_group is not None and path in h5_group:
                    return h5py_read_dataset(h5_group[path])
            if message is not None:
                getattr(_logger, level)(message)

        if entry is None:
            if url is None:
                return None
            with EntryReader(url) as h5_group:
                return process(h5_group)
        else:
            return process(entry)

    def _get_source_name(self):
        """ """
        return self._generic_path_getter(
            paths=self._SOURCE_NAME, message="Unable to find source name", level="info"
        )

    def _get_source_type(self):
        """ """
        return self._generic_path_getter(
            paths=self._SOURCE_TYPE, message="Unable to find source type", level="info"
        )

    def _get_title(self):
        """return acquisition title"""
        return self._generic_path_getter(
            paths=self.TITLE_PATHS, message="Unable to find title"
        )

    def _get_instrument_name(self):
        """:return instrument instrument name (aka beamline name)"""
        name = self._generic_path_getter(
            paths=self._INSTRUMENT_NAME_PATH,
            message="Unable to find instrument name",
            level="info",
        )
        # on some path / old hdf5 the name is prefixed by "ESRF:". clean those
        if name is not None and name.startswith("ESRF:"):
            name = name.replace("ESRF:", "")
        return name

    def _get_dataset_name(self):
        """return name of the acquisition"""
        return self._generic_path_getter(
            paths=self._DATASET_NAME_PATH,
            message="No name describing the acquisition has been "
            "found, Name dataset will be skip",
        )

    def _get_sample_name(self):
        """return sample name"""
        return self._generic_path_getter(
            paths=self._SAMPLE_NAME_PATH,
            message="No sample name has been "
            "found, Sample name dataset will be skip",
        )

    def _get_grp_size(self):
        """return the nb_scans composing the zseries if is part of a group
        of sequence"""
        return self._generic_path_getter(paths=self._GRP_SIZE_PATH, message=None)

    def _get_tomo_n(self):
        return self._generic_path_getter(
            paths=self._TOMO_N_PATH,
            message="unable to find information regarding tomo_n",
        )

    def _get_start_time(self, entry=None):
        return self._generic_path_getter(
            paths=self._START_TIME_PATH,
            message="Unable to find start time",
            level="info",
            entry=entry,
        )

    def _get_end_time(self, entry=None):
        return self._generic_path_getter(
            paths=self._END_TIME_PATH,
            message="Unable to find end time",
            level="info",
            entry=entry,
        )

    def _get_soft_flip_frame(self):
        """
        Retrieve registered frame flip (know by bliss-tomo)
        """
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None, None
        self._check_has_metadata(url)
        with EntryReader(url) as entry:
            for flip_path in self._FRAME_FLIP_PATHS:
                if len(self._unique_detector_names) > 0:
                    key = flip_path.format(detector_name=self._unique_detector_names[0])
                else:
                    key = flip_path
                if key in entry:
                    return h5py_read_dataset(entry[key])
            else:
                return None, None

    def _get_propagation_distance(self) -> pint.Quantity:
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None
        self._check_has_metadata(url)
        with EntryReader(url) as entry:
            for prog_dst_path in self._PROPAGATION_DISTANCE_PATHS:
                if prog_dst_path in entry:
                    dataset = entry[prog_dst_path]
                    prog_dst = h5py_read_dataset(dataset)
                    unit = get_dataset_unit(
                        dataset,
                        default=_ureg.mm,
                        from_dataset=f"{dataset.name} (reading propagation distance)",
                    )
                    return prog_dst * unit
            else:
                return None

    def _get_energy(self, ask_if_0, input_callback) -> pint.Quantity | None:
        """Try to read the energy from root url.
        If fails and if a input_callback given then execute this fallback
        """
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None
        self._check_has_metadata()
        with EntryReader(url) as entry:
            if self._ENERGY_PATH in entry:
                dataset = entry[self._ENERGY_PATH]
                energy = h5py_read_dataset(dataset)
                unit = get_dataset_unit(
                    dataset,
                    default=_ureg.keV,
                    from_dataset=f"{dataset.name} (from reading energy)",
                )
                if energy == 0 and ask_if_0:
                    desc = (
                        "Energy has not been registered. Please enter "
                        "incoming beam energy (in kev):"
                    )
                    if input_callback is None:
                        en = input(desc)
                    else:
                        en = input_callback("energy", desc)
                    if energy is not None:
                        energy = float(en)
                return energy * unit
            else:
                mess = f"unable to find energy for entry {entry}."
                if self.raise_error_if_issue:
                    raise ValueError(mess)
                else:
                    mess += " Default value will be set (19kev)"
                    _logger.warning(mess)
                    return 19.0 * _ureg.keV

    def _get_sample_detector_distance(self) -> pint.Quantity | None:
        """return tuple(distance, unit)"""
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None
        self._check_has_metadata(url)
        with EntryReader(url) as entry:
            for key in self.configuration.sample_detector_distance_keys:
                if key in entry:
                    dataset = entry[key]
                    distance = h5py_read_dataset(dataset)
                    unit = get_dataset_unit(
                        dataset,
                        default=_ureg.mm,
                        from_dataset=f"{dataset.name} (from reading sample-detector distance)",
                    )
                    # convert to meter
                    return distance * unit

            mess = f"unable to find distance for entry {entry}."
            if self.raise_error_if_issue:
                raise ValueError(mess)
            else:
                mess += "Default value will be set (1m)"
                _logger.warning(mess)
                return 1.0 * _ureg.meter

    def _get_source_sample_distance(self) -> pint.Quantity | None:
        """return tuple(distance, unit)"""
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None
        self._check_has_metadata(url)
        with EntryReader(url) as entry:
            for key in self.configuration.source_sample_distance_keys:
                if key in entry:
                    dataset = entry[key]
                    distance = h5py_read_dataset(dataset)
                    unit = get_dataset_unit(
                        dataset,
                        default=_ureg.mm,
                        from_dataset=f"{dataset.name} (from reading source-sample distance)",
                    )
                    # convert to meter
                    return distance * unit

            mess = f"unable to find distance for entry {entry}."
            if self.raise_error_if_issue:
                raise ValueError(mess)
            else:
                mess += "Default value will be set (1m)"
                _logger.warning(mess)
                return 1.0 * _ureg.meter

    def _get_sample_pixel_size(self, axis) -> pint.Quantity | None:
        """read **sample** pixel size from predefined set of path"""
        return self._get_pixel_size(
            axis=axis,
            x_pixel_size_paths=self.configuration.sample_x_pixel_size_keys,
            y_pixel_size_paths=self.configuration.sample_y_pixel_size_keys,
        )

    def _get_detector_pixel_size(self, axis) -> pint.Quantity | None:
        """read **detector** pixel size from predefined set of path"""
        return self._get_pixel_size(
            axis=axis,
            x_pixel_size_paths=self.configuration.detector_x_pixel_size_keys,
            y_pixel_size_paths=self.configuration.detector_y_pixel_size_keys,
        )

    def _get_pixel_size(self, axis, x_pixel_size_paths, y_pixel_size_paths):
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None
        if axis not in ("x", "y"):
            raise ValueError
        self._check_has_metadata()

        if axis == "x":
            keys = x_pixel_size_paths
        elif axis == "y":
            keys = y_pixel_size_paths
        else:
            raise ValueError(f"axis {axis} is invalid")

        # solve according to detector name
        if len(self._unique_detector_names) > 1:
            _logger.warning(
                f"More than one detector found. Will pick the first one found ({self._unique_detector_names[0]})"
            )
        if len(self._unique_detector_names) > 0:
            keys = [
                key.format(detector_name=self._unique_detector_names[0]) for key in keys
            ]

        self._unique_detector_names

        with EntryReader(url) as entry:
            for key in keys:
                if key in entry:
                    dataset = entry[key]
                    dataset_value = h5py_read_dataset(dataset)
                    # if the pixel size is provided as x, y
                    if isinstance(dataset_value, numpy.ndarray):
                        if len(dataset_value) > 1 and axis == "y":
                            size_ = dataset_value[1]
                        else:
                            size_ = dataset_value[0]
                    # if this is a single value
                    else:
                        size_ = dataset_value
                    unit = get_dataset_unit(
                        dataset,
                        default=_ureg.micrometer,
                        from_dataset=f"{dataset.name} (from reading pixel size)",
                    )
                    # convert to meter
                    return size_ * unit

            mess = f"unable to find {axis} sample pixel size for entry {entry}"
            if self.raise_error_if_issue:
                raise ValueError(mess)
            else:
                mess += "Value will be set to default (10-6m)"
                _logger.warning(mess)
                return 10e-6 * _ureg.meter

    def _get_field_of_view(self) -> str:
        if self.configuration.field_of_view is not None:
            return self.configuration.field_of_view.value
        url = self.parent_root_url() or self.root_url
        if url is None:
            return None
        with EntryReader(url) as entry:
            if self._FOV_PATH in entry:
                return h5py_read_dataset(entry[self._FOV_PATH])
            else:
                # FOV is optional: don't raise an error
                _logger.warning(
                    f"unable to find information regarding field of view for entry {entry}. set it to default value (Full)"
                )
                return "Full"

    def _update_configuration_from_tomo_config(self):
        """
        force some values from EBS tomo 'tomoconfig' group to make sure correct dataset are read
        """
        if self.configuration.ignore_bliss_tomo_config:
            return
        url = self.parent_root_url() or self.root_url
        if url is None:
            # case of entries are made manually and user do not provide an init node.
            return
        with EntryReader(url) as entry:
            technique_grp = entry.get("technique", None)
            if technique_grp is None:
                _logger.warning(
                    f"Unable to find a technique group in {entry}. Unable to reach EBStomo metadata"
                )
                return

            bliss_tomo_version = technique_grp.attrs.get("tomo_version", None)

            # read metadata
            try:
                bliss_metadata = BlissTomoConfig.from_technique_group(
                    technique_group=technique_grp
                )
            except KeyError:
                if bliss_tomo_version is not None:
                    _logger.warning(
                        f"Unable to find bliss 'tomo_config' when expected (tomo_version={bliss_tomo_version}). Fallback to conversion based on list of paths to check"
                    )
            else:
                # check if some metadata are missing
                metadata_values = {
                    "detector": bliss_metadata.tomo_detector,
                    "sample_x": bliss_metadata.sample_x,
                    "sample_y": bliss_metadata.sample_y,
                    "translation_z": bliss_metadata.translation_z,
                    "rotation": bliss_metadata.rotation,
                    "rotation_is_clockwise": bliss_metadata.rotation_is_clockwise,
                }
                missing_metadata = list(
                    [k for k, v in metadata_values.items() if v is None]
                )
                _logger.info(f"read tomo config from bliss. Get {metadata_values}")
                if len(missing_metadata) > 0:
                    _logger.warning(
                        f"couldn't find {missing_metadata} in bliss 'technique/tomoconfig' dataset"
                    )

                if bliss_metadata.tomo_detector is not None:
                    self.configuration.valid_camera_names = bliss_metadata.tomo_detector
                if bliss_metadata.sample_x is not None:
                    self.configuration.sample_x_keys = bliss_metadata.sample_x
                if bliss_metadata.sample_y is not None:
                    self.configuration.sample_y_keys = bliss_metadata.sample_y
                if bliss_metadata.translation_y is not None:
                    self.configuration.translation_y_keys = bliss_metadata.translation_y
                if bliss_metadata.translation_z is not None:
                    self.configuration.translation_z_keys = bliss_metadata.translation_z
                if bliss_metadata.rotation is not None:
                    self.configuration.rotation_angle_keys = bliss_metadata.rotation
                # if the 'rotation_is_clockwise' has not been defined by the user and if we can access the
                # metadata from blisstomo take it.
                if (
                    bliss_metadata.rotation_is_clockwise is not None
                    and self.configuration.rotation_is_clockwise is None
                ):
                    self.configuration.rotation_is_clockwise = (
                        bliss_metadata.rotation_is_clockwise
                    )

    def to_NXtomos(self, request_input, input_callback, check_tomo_n=True) -> tuple:
        self._update_configuration_from_tomo_config()
        self._preprocess_registered_entries()

        nx_tomo = NXtomo()

        # 1. root level information
        # start and end time
        nx_tomo.start_time = self._get_start_time()
        nx_tomo.end_time = self._get_end_time()

        # title
        nx_tomo.title = self._get_dataset_name()
        # group size
        nx_tomo.group_size = self._get_grp_size()

        # 2. define beam
        try:
            energy: pint.Quantity = self._get_user_settable_parameter(
                param_key="energy_kev",
                fallback_fct=self._get_energy,
                input_callback=input_callback,
                ask_if_0=request_input,
            )
        except TypeError as e:
            # if the path lead to unexpected dataset (group instead of a dataset or a broken folder)
            _logger.error("Fail to get energy. Error is %s", e, exc_info=e)
            energy = None

        if energy is not None:
            # TODO: better management of energy ? might be energy.beam or energy.instrument.beam ?
            nx_tomo.energy = energy

        # 3. define instrument
        nx_tomo.instrument.name = self._get_instrument_name()
        nx_tomo.instrument.detector.data = self._virtual_sources
        nx_tomo.instrument.detector.image_key_control = self.image_key_control
        nx_tomo.instrument.detector.count_time = self._acq_expo_time
        nx_tomo.instrument.detector.roi = self.get_detector_roi()
        if self.image_key_control is None:
            _logger.warning(
                "No image key defined. Unable to determine the number of frame and the type. This is a mandatory field"
            )
        else:
            n_frames = len(self.image_key_control)
            nx_tomo.instrument.detector.sequence_number = numpy.linspace(
                start=0, stop=n_frames, num=n_frames, dtype=numpy.uint32, endpoint=False
            )

        # sample - detector distance
        try:
            sample_detector_distance = self._get_user_settable_parameter(
                param_key="detector_sample_distance_m",
                fallback_fct=self._get_sample_detector_distance,
            )
        except TypeError:
            # if the path lead to unexpected dataset (group instead of a dataset or a broken folder)
            _logger.error("Fail to get sample/detector distance")
            sample_detector_distance = None
        if sample_detector_distance is not None:
            nx_tomo.instrument.detector.distance = sample_detector_distance

        # source - sample detector distance
        try:
            source_sample_distance: pint.Quantity = self._get_user_settable_parameter(
                param_key="source_sample_distance_m",
                fallback_fct=self._get_source_sample_distance,
            )
        except TypeError:
            # if the path lead to unexpected dataset (group instead of a dataset or a broken folder)
            _logger.error("Fail to get source/sample distance")
            source_sample_distance = None
        if source_sample_distance is not None:
            # source_sample_distance is positive when the NXtomo application states it should be negative. So
            # let's make sure this will be negative
            source_sample_distance = -numpy.abs(source_sample_distance)
            nx_tomo.instrument.source.distance = source_sample_distance

        # handle detector & sample ; x & y pixel size
        for (nx_obj, attr), params in {
            (nx_tomo.sample, "x_pixel_size"): {
                "param_key": "x_sample_pixel_size_m",
                "fallback_fct": self._get_sample_pixel_size,
                "axis": "x",
            },
            (nx_tomo.sample, "y_pixel_size"): {
                "param_key": "y_sample_pixel_size_m",
                "fallback_fct": self._get_sample_pixel_size,
                "axis": "y",
            },
            (nx_tomo.instrument.detector, "x_pixel_size"): {
                "param_key": "x_detector_pixel_size_m",
                "fallback_fct": self._get_detector_pixel_size,
                "axis": "x",
            },
            (nx_tomo.instrument.detector, "y_pixel_size"): {
                "param_key": "y_detector_pixel_size_m",
                "fallback_fct": self._get_detector_pixel_size,
                "axis": "y",
            },
        }.items():
            try:
                pixel_size = self._get_user_settable_parameter(**params)
            except TypeError:
                # if the path lead to unexpected dataset (group instead of a dataset or a broken folder)
                _logger.error(f"Fail to get {attr} of {nx_obj}")
                pixel_size = None
            else:
                nx_obj.__setattr__(attr, pixel_size)

        # flips
        nx_tomo.instrument.detector.transformations.add_transformation(
            DetYFlipTransformation(flip=self.lr_flipped)
        )
        nx_tomo.instrument.detector.transformations.add_transformation(
            DetXFlipTransformation(flip=self.ud_flipped)
        )

        # fov
        fov = self._get_field_of_view()
        if fov is not None:
            nx_tomo.instrument.detector.field_of_view = fov

        # x_rotation_axis_pixel_position
        # TODO Missing calibration mechanism from Bliss
        if self.translation_y is None or len(self.translation_y) < 1:
            _logger.warning("Unable to find translation_y")
        elif nx_tomo.sample.x_pixel_size is not None:
            x_sample_pixel_size = nx_tomo.sample.x_pixel_size
            translation_y_m = (
                self.translation_y[0].to(_ureg.meter).magnitude
            )  # This is very fragile, no idea how to make it better
            x_sample_pixel_size = x_sample_pixel_size.to(_ureg.meter).magnitude
            nx_tomo.instrument.detector.x_rotation_axis_pixel_position = (
                translation_y_m / x_sample_pixel_size
            )

        # define tomo_n
        nx_tomo.instrument.detector.tomo_n = self._get_tomo_n()

        # 4. define nx source
        source_name = self._get_source_name()
        nx_tomo.instrument.source.name = source_name
        source_type = self._get_source_type()
        if source_type is not None:
            if "synchrotron" in source_type.lower():
                source_type = SourceType.SYNCHROTRON_X_RAY_SOURCE.value
            # drop a warning if the source type is invalid
            if source_type not in SourceType.values():
                _logger.warning(
                    f"Source type ({source_type}) is not a 'standard value'"
                )

        nx_tomo.instrument.source.type = source_type

        # 5. define sample
        # warning: the NXtomo XYZ reference (see https://manual.nexusformat.org/_static/NeXusManual.pdf - p 26)
        #
        #
        #                    Y axis
        #                      ^   X axis
        #                      |  /
        #    x-ray             | /
        #    -------->          ------> Z axis
        #
        # is not the same as the ESRF XYZ (see https://tomo.gitlab-pages.esrf.fr/bliss-tomo/master/modelization_sample_stage.html)
        #
        #               Z axis
        #                      ^    Y axis
        #                      | /
        #    x-ray             |/
        #    -------->          ------> X axis
        #
        #
        nx_tomo.sample.name = self._get_sample_name()
        assert isinstance(self.rotation_angle, (pint.Quantity, type(None)))
        nx_tomo.sample.rotation_angle = self.rotation_angle
        assert isinstance(self.sample_x, (pint.Quantity, type(None)))
        nx_tomo.sample.z_translation = self.sample_x
        assert isinstance(self.translation_z, (pint.Quantity, type(None)))
        nx_tomo.sample.y_translation = self.translation_z
        assert isinstance(self.sample_y, (pint.Quantity, type(None)))
        nx_tomo.sample.x_translation = self.sample_y

        z1 = nx_tomo.instrument.source.distance
        z2 = nx_tomo.instrument.detector.distance
        propagation_distance: pint.Quantity | None = self._get_propagation_distance()
        if propagation_distance is not None:
            _logger.debug("set propagation distance from bliss metadata")
            nx_tomo.sample.propagation_distance = propagation_distance
        elif z1 is not None and z2 is not None:
            _logger.debug(
                "compute propagation distance from sample-source and sample-detector distances"
            )
            nx_tomo.sample.propagation_distance = (-z1 * z2) / (-z1 + z2)
        else:
            distances = {
                "sample-detector distance": z1,
                "source-sample distance": z2,
            }
            _logger.warning(
                "Unable to define propagation distance. %s missing",
                [key for key, value in distances.items() if value is not None],
            )

        # 6. define control
        if (
            self.configuration.create_control_data
            and self.known_machine_current not in (None, dict())
        ):
            nx_tomo.control.data = (
                deduce_machine_current(
                    timestamps=self._frames_timestamp,
                    known_machine_current=self._known_machine_current_am,
                )
                * _ureg.ampere
            )
            types = set()
            if nx_tomo.control.data is not None:
                for d in nx_tomo.control.data:
                    types.add(type(d))

        if check_tomo_n:
            self.check_tomo_n()
        return (nx_tomo,)

    def check_tomo_n(self):
        # check scan is complete
        tomo_n = self._get_tomo_n()
        if self.configuration.check_tomo_n and tomo_n is not None:
            image_key_control = numpy.asarray(self._image_key_control)
            proj_found = len(
                image_key_control[image_key_control == ImageKey.PROJECTION.value]
            )
            if proj_found < tomo_n:
                mess = f"Incomplete scan. Expect {tomo_n} projection but only {proj_found} found"
                if self.configuration.raises_error is True:
                    raise ValueError(mess)
                else:
                    _logger.error(mess)

    def _check_has_metadata(self, url: DataUrl | None = None):
        url = url or self.root_url
        if url is None:
            raise ValueError(
                "no initialization entry specify, unable to" "retrieve energy"
            )

    def _get_user_settable_parameter(
        self,
        param_key,
        fallback_fct,
        *fallback_args,
        **fallback_kwargs,
    ) -> pint.Quantity | None:
        """
        return value, unit

        :param fallback_fct: callback function to retrieve the value. Must return a quantity
        """
        value = getattr(self.configuration, param_key, None)
        if value is not None:
            units = {
                "energy_kev": _ureg.keV,
                "x_sample_pixel_size_m": _ureg.meter,
                "y_sample_pixel_size_m": _ureg.meter,
                "x_detector_pixel_size_m": _ureg.meter,
                "y_detector_pixel_size_m": _ureg.meter,
                "detector_sample_distance_m": _ureg.meter,
                "source_sample_distance_m": _ureg.meter,
            }
            if param_key not in units:
                raise ValueError(f"key {param_key} is not handled")
            unit = units[param_key]
            value = value * unit
        else:
            value = fallback_fct(*fallback_args, **fallback_kwargs)
            assert value is None or isinstance(
                value, pint.Quantity
            ), "fallback must return a pint.Quantity"

        return value



# coding: utf-8
"""
module to define a multi-tomo acquistion (also know as pco-tomo)
"""

from __future__ import annotations

import logging

from silx.io.url import DataUrl
from silx.utils.proxy import docstring
from silx.utils.deprecation import deprecated

from nxtomo.application.nxtomo import NXtomo

try:
    from nxtomo.utils.NXtomoSplitter import NXtomoSplitter
except ImportError:
    from nxtomo.utils.detectorsplitter import (
        NXtomoDetectorDataSplitter as NXtomoSplitter,
    )
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.converter.hdf5.acquisition.baseacquisition import EntryReader
from nxtomomill.converter.hdf5.acquisition.standardacquisition import (
    StandardAcquisition,
)
from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.io.config import TomoHDF5Config

_logger = logging.getLogger(__name__)

__all__ = [
    "MultiTomoAcquisition",
]


class MultiTomoAcquisition(StandardAcquisition):
    """
    A multitomo is an acquisition that we want to split into several NXtomos.

    It is contain in a bliss scan and looks like the following at bliss side:

    * scan "description", title can be tomo:pcotomo ro tomo:multitomo for example
    * Optional scan dark
    * Optional scan flat
    * scan "projections"
    * Optional scan flat

    The scan "projections" contains several "tomo" with a parameter evolving with time like heat or pressure.
    The idea is that we want to split those into several NXtomo.
    Those NXtomo must duplicate dark and flat scans
    For example if scan "projections" contains nb_loop = 2 and nb_loop = 3 we must create nb_loop*nb_loop == 6 NXtomo as output.

    Split of those can are done in postprocessing on the to_NXtomos function
    """

    def __init__(
        self,
        root_url: DataUrl | None,
        configuration: TomoHDF5Config,
        detector_sel_callback,
        start_index,
    ):
        super().__init__(
            root_url=root_url,
            configuration=configuration,
            detector_sel_callback=detector_sel_callback,
            start_index=start_index,
        )
        self._nb_loop = None
        self._nb_tomo = None
        self._nb_turns = None
        self._entries_to_split = set()  # set of URL as str

    def _preprocess_registered_entries(self):
        # those must be defined before calling super because the super will call then `_preprocess_registered_entry`
        self._nb_loop = None
        self._nb_tomo = None
        self._nb_turns = None
        self._entries_to_split = set()
        super()._preprocess_registered_entries()

    def _read_nb_loop(self, entry_url: DataUrl):
        return self.get_nb_loop(entry_url) or self.get_nb_loop(self.root_url)

    def _read_nb_tomo(self, entry_url: DataUrl):
        return self.get_nb_tomo(entry_url) or self.get_nb_tomo(self.root_url)

    def _read_nb_turns(self, entry_url: DataUrl):
        return self.get_nb_turns(entry_url) or self.get_nb_turns(self.root_url)

    def _preprocess_registered_entry(self, entry_url, type_):
        super()._preprocess_registered_entry(entry_url=entry_url, type_=type_)
        if type_ is AcquisitionStep.PROJECTION:
            # nb loop parameter must be present only on projection entries
            nb_loop = self._read_nb_loop(entry_url=entry_url)
            if (
                nb_loop is not None
            ):  # at this moment 02/2022 nb_loop is only defined on projection type
                if self._nb_loop is None or self._nb_loop == nb_loop:
                    self._nb_loop = nb_loop
                    self._entries_to_split.add(entry_url.path())
                else:
                    _logger.error(
                        f"Found entries with a different number of nb_loop: {entry_url.path()}"
                    )

            nb_tomo = self._read_nb_tomo(entry_url=entry_url)
            if (
                nb_tomo is not None
            ):  # at this moment 02/2022 nb_loop is only defined on projection type
                if self._nb_tomo is None or self._nb_tomo == nb_tomo:
                    self._nb_tomo = nb_tomo
                    self._entries_to_split.add(entry_url.path())
                else:
                    _logger.error(
                        f"Found entries with a different number of _nb_tomo: {entry_url.path()}"
                    )

            nb_turns = self._read_nb_turns(entry_url=entry_url)
            if nb_turns is not None:
                if self._nb_turns is None or self._nb_turns == nb_turns:
                    self._nb_turns = nb_turns
                    self._entries_to_split.add(entry_url.path())
                else:
                    _logger.error(
                        f"Found entries with a different number of nb_turns: {entry_url.path()}"
                    )

    def get_nb_loop(self, url) -> int | None:
        with EntryReader(url) as entry:
            for path in self._NB_LOOP_PATH:
                if path in entry:
                    return entry[path][()]
        return None

    def get_nb_tomo(self, url) -> int | None:
        with EntryReader(url) as entry:
            for path in self._NB_TOMO_PATH:
                if path in entry:
                    return entry[path][()]
        return None

    def get_nb_turns(self, url) -> int | None:
        with EntryReader(url) as entry:
            for path in self._NB_TURNS_PATH:
                if path in entry:
                    return entry[path][()]
        return None

    def get_expected_nx_tomo(self):
        # the number of expected NXtomo is saved with projection
        # and not with the init title. This is why it but be computed later
        return 0

    @deprecated(
        replacement="get_multitomo_version",
        reason="rename pcotomo to multitomo",
        since_version="1.2",
    )
    def get_pcotomo_version(self, url: DataUrl):
        return self.get_multitomo_version(url=url)

    def get_multitomo_version(self, url: DataUrl):
        """
        return the multitomo version according to the provider version (aka bliss)
        """
        with EntryReader(
            url=DataUrl(file_path=url.file_path(), data_path="/", scheme=url.scheme())
        ) as entry:
            if "creator_version" in entry.attrs:
                creator_version = entry.attrs["creator_version"]
                try:
                    full_version = creator_version.split("-")[0].split("+")[0]
                    major, minor, _ = full_version.split(".")
                except Exception as e:
                    _logger.warning(
                        f"Fail to convert creator_version ({creator_version}) to a valid version number. Error is {e}"
                    )
                    return None
                else:
                    if major == "0" or (major == "1" and int(minor) < 9):
                        return 1
                    else:
                        return 2
        return None

    @docstring(StandardAcquisition)
    def to_NXtomos(
        self, request_input, input_callback, check_tomo_n: bool = True
    ) -> tuple:
        multitomo_version = self.get_multitomo_version(url=self.root_url)
        if multitomo_version is None:
            multitomo_version = 2
            _logger.warning(
                f"No multi-tomo (also know as pco-tomo) version found. Try to convert it using the latest version (v{multitomo_version})"
            )
        if multitomo_version == 1:
            return self.to_NXtomos_multitomo_v1(
                request_input=request_input,
                input_callback=input_callback,
                check_tomo_n=check_tomo_n,
            )
        elif multitomo_version == 2:
            return self.to_NXtomos_multitomo_v2(
                request_input=request_input,
                input_callback=input_callback,
                check_tomo_n=check_tomo_n,
            )
        else:
            raise ValueError(
                f"{multitomo_version} version of the multi-tomo found. No handling defined for it"
            )

    @deprecated(
        replacement="to_NXtomos_multitomo_v2",
        reason="rename pcotomo to multitomo",
        since_version="1.2",
    )
    def to_NXtomos_pcotomo_v2(
        self, request_input, input_callback, check_tomo_n: bool = True
    ) -> tuple:
        return self.to_NXtomos_multitomo_v2(
            request_input=request_input,
            input_callback=input_callback,
            check_tomo_n=check_tomo_n,
        )

    def to_NXtomos_multitomo_v2(
        self, request_input, input_callback, check_tomo_n: bool = True
    ) -> tuple:
        """
        The second version of 'to_NXtomos' for the second way of storing multi-tomo information.
        In this version we expect to have:
        * nb_turns: number of NXtomo to generate
        * proj_n: number of projections per NXtomo
        * scan_range: rotation angle scope

        """
        nx_tomos = super().to_NXtomos(request_input, input_callback, check_tomo_n=False)
        if len(nx_tomos) == 0:
            return nx_tomos

        if self._nb_turns is None:
            error_msg = "unable to find nb_turns information to split the NXtomo"
            _logger.error(error_msg)
            raise ValueError(error_msg)

        # if we want to split the NXtomo from `nb_tomo` and `nb_loop`
        _logger.info(
            "apply split from `nb_turns` and `scan_range`. `start_angle_offset`, `angle_interval` and `n_tomo` will be ignored"
        )

        results = []
        for nx_tomo in nx_tomos:
            splitter = NXtomoSplitter(nx_tomo)
            projections_slices = self._get_projections_slices(nx_tomo)
            if len(projections_slices) > 1:
                # insure projections are contiguous otherwise we don't know how to split it.
                # not defined on the current design from bliss. should never happen
                raise ValueError("Expect all projections to be contiguous")
            elif len(projections_slices) == 0:
                raise ValueError("No projection found")
            else:
                results.extend(
                    splitter.split(
                        projections_slices[0],
                        nb_part=int(self._nb_turns),
                        tomo_n=self._get_tomo_n(),
                    )
                )

        if check_tomo_n:
            # if angle offset provided there is an hight probably this will mess up with the tomo_n, which is expected...
            self.check_tomo_n()

        return tuple(results)

    @deprecated(
        replacement="to_NXtomos_multitomo_v1",
        reason="rename pcotomo to multitomo",
        since_version="1.2",
    )
    def to_NXtomos_pcotomo_v1(
        self, request_input, input_callback, check_tomo_n: bool = True
    ) -> tuple:
        return self.to_NXtomos_multitomo_v1(
            request_input=request_input,
            input_callback=input_callback,
            check_tomo_n=check_tomo_n,
        )

    def to_NXtomos_multitomo_v1(
        self, request_input, input_callback, check_tomo_n: bool = True
    ) -> tuple:
        """
        The first version of 'to_NXtomos' for the first way of storing multitomo information.
        In this version we expect to have:
        * `nb_loop`: define the number of turn
        * `nb_tomo`: define the number of 'sequence' per turn. One sequence will generate one NXtomo

        """

        nx_tomos = super().to_NXtomos(request_input, input_callback, check_tomo_n=False)
        if len(nx_tomos) == 0:
            return nx_tomos

        # apply multitomo tomo_n, angle_interval and start_angle_offset
        # this will select if necessary a subset of the NXtomo to apply user requested offset
        angle_offset = self.configuration.multitomo_start_angle_offset
        scan_range = self.configuration.multitomo_scan_range
        if scan_range not in (180, 360):
            _logger.warning(
                f"strange angle_interval provided: {scan_range} when 180 or 360 expected"
            )
        results = []
        if angle_offset is not None:
            # if we want to split the NXtomo from angles and settings provided by the user:
            _logger.info(
                "apply filtering from `start_angle_offset`, `angle_interval` and `n_tomo`. `nb_tomo` and `nb_loop` will be ignored"
            )
            for nx_tomo in nx_tomos:
                for i_part in range(self.configuration.multitomo_n_nxtomo):
                    start_angle_offset = (angle_offset or 0) + (i_part * scan_range)
                    results.append(
                        NXtomo.sub_select_from_angle_offset(
                            nx_tomo=nx_tomo,
                            start_angle_offset=start_angle_offset,
                            angle_interval=scan_range,
                            shift_angles=False,
                            copy=True,
                        )
                    )
        else:
            # if we want to split the NXtomo from `nb_tomo` and `nb_loop`
            _logger.info(
                "apply split from `nb_tomo` and `nb_loop`. `start_angle_offset`, `angle_interval` and `n_tomo` will be ignored"
            )

            for nx_tomo in nx_tomos:
                splitter = NXtomoSplitter(nx_tomo)
                projections_slices = self._get_projections_slices(nx_tomo)
                if len(projections_slices) > 1:
                    # insure projections are contiguous otherwise we don't know how to split it.
                    # not defined on the current design from bliss. should never happen
                    raise ValueError("Expect all projections to be contiguous")
                elif len(projections_slices) == 0:
                    raise ValueError("No projection found")
                else:
                    results.extend(
                        splitter.split(
                            projections_slices[0],
                            nb_part=int(self._nb_loop * self._nb_tomo),
                        )
                    )

        if self.configuration.multitomo_shift_angles:
            _logger.info(f"will shift angle to 0-{scan_range}")
            results = [
                NXtomo.clamp_angles(
                    nx_tomo,
                    angle_range=scan_range,
                    offset=angle_offset or 0,
                    copy=False,
                    image_keys=(ImageKey.PROJECTION, ImageKey.ALIGNMENT),
                )
                for nx_tomo in results
            ]

        if angle_offset is not None and check_tomo_n:
            # if angle offset provided there is an hight probably this will mess up with the tomo_n, which is expected...
            self.check_tomo_n()

        return tuple(results)

    @staticmethod
    def _get_projections_slices(nx_tomo: NXtomo) -> tuple:
        """Return a tuple of slices for each group of contiguous projections"""
        if nx_tomo.instrument.detector.image_key_control is None:
            return ()

        res = []
        start_pos = -1
        browsing_projection = False
        for i_frame, image_key in enumerate(
            nx_tomo.instrument.detector.image_key_control
        ):
            image_key_value = ImageKey(image_key)
            if image_key_value is ImageKey.PROJECTION and not browsing_projection:
                browsing_projection = True
                start_pos = i_frame
            elif browsing_projection and image_key_value is not ImageKey.PROJECTION:
                res.append(slice(start_pos, i_frame, 1))
                start_pos = -1
                browsing_projection = False
        else:
            if browsing_projection is True:
                res.append(slice(start_pos, i_frame + 1, 1))
        return tuple(res)



# coding: utf-8

import os
import tempfile

from nxtomomill.app.dxfile2nx import main
from nxtomomill.tests.utils.dxfile import MockDxFile


def test_dxfile2nx_application():
    """test nxtomomill dxfile2nx input_file output_file"""
    with tempfile.TemporaryDirectory() as folder:
        dx_file_path = os.path.join(folder, "dxfile.dx")
        nx_file_path = os.path.join(folder, "dxfile.nx")
        n_projections = 50
        n_darks = 2
        n_flats = 4

        MockDxFile(
            file_path=dx_file_path,
            n_projection=n_projections,
            n_darks=n_darks,
            n_flats=n_flats,
        )
        assert not os.path.exists(nx_file_path), "outputfile exists already"
        main(["dxfile2nx", dx_file_path, nx_file_path])
        assert os.path.exists(nx_file_path), "outputfile doesn't exists"



from __future__ import annotations

import logging
from pydantic import BaseModel, Field, field_validator, field_serializer, ConfigDict
from nxtomomill.utils import FileExtension
from nxtomomill.models.utils import filter_str_def


class GeneralSection(BaseModel):
    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    output_file: str | None = Field(default=None, description="Output file name")
    file_extension: FileExtension = Field(
        default=FileExtension.NX,
        description="File extension if not given in the output file name",
    )
    overwrite: bool = Field(
        default=False, description="Overwrite output files if exists"
    )
    log_level: int = Field(
        default=logging.WARNING,
        description='Log level: "debug", "info", "warning", "error"',
    )

    @field_validator(
        "log_level",
        mode="before",
    )
    @classmethod
    def cast_to_log_level(cls, value: str | int) -> int:
        if isinstance(value, int):
            return value
        elif isinstance(value, str):
            value = filter_str_def(value)
            return getattr(logging, value.upper())
        else:
            raise TypeError

    @field_validator(
        "file_extension",
        mode="plain",
    )
    @classmethod
    def cast_to_file_extension(cls, value: str | FileExtension) -> FileExtension:
        if isinstance(value, str):
            value = filter_str_def(value)
        return FileExtension(value)

    @field_serializer(
        "log_level",
        when_used="always",
    )
    @classmethod
    def serialize_log_level(cls, value: int) -> str:
        return logging.getLevelName(value).lower()

    @field_serializer(
        "file_extension",
        when_used="always",
    )
    @classmethod
    def serialize_file_extension(cls, file_ext: FileExtension) -> str:
        return file_ext.value

    @field_validator(
        "output_file",
    )
    @classmethod
    def cast_output_file(cls, output_file: str | None) -> str | None:
        return filter_str_def(output_file)



# coding: utf-8

import os
import pytest
import numpy

from nxtomomill import converter
from nxtomomill.tests.utils.dxfile import MockDxFile

from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.validator import is_valid_for_reconstruction

from silx.io.utils import get_data


@pytest.mark.parametrize("duplicate_data", (True, False))
def test_dx_file_conversion(tmp_path, duplicate_data):

    folder = tmp_path / "test_dx_file_conversion"
    folder.mkdir()
    dxfile_path = os.path.join(folder, "dxfile.h5")

    n_projections = 50
    n_darks = 2
    n_flats = 4
    mock = MockDxFile(
        file_path=dxfile_path,
        n_projection=n_projections,
        n_darks=n_darks,
        n_flats=n_flats,
    )

    output_file = os.path.join(folder, "dxfile.nx")

    results = converter.from_dx_to_nx(
        input_file=dxfile_path,
        output_file=output_file,
        duplicate_data=duplicate_data,
    )
    assert len(results) == 1
    assert os.path.exists(output_file)
    _, entry = results[0]
    scan = NXtomoScan(output_file, entry)
    assert len(scan.projections) == n_projections
    assert len(scan.darks) == n_darks
    assert len(scan.flats) == n_flats
    assert numpy.array(scan.rotation_angle).min() == 0
    assert numpy.array(scan.rotation_angle).max() == 180
    assert is_valid_for_reconstruction(scan)

    # check arrays are correctly copied from mock
    numpy.testing.assert_array_equal(mock.data_dark[0], get_data(scan.darks[0]))
    numpy.testing.assert_array_equal(mock.data_flat[1], get_data(scan.flats[3]))
    idx_last_proj = n_projections + n_flats + n_darks - 1
    numpy.testing.assert_array_equal(
        mock.data_proj[-1], get_data(scan.projections[idx_last_proj])
    )
    assert scan.rotation_angle[0] == 0  # pylint: disable=E1136
    assert scan.rotation_angle[5] == 0  # pylint: disable=E1136
    assert scan.rotation_angle[6] == 0  # pylint: disable=E1136
    assert scan.rotation_angle[-1] == 180  # pylint: disable=E1136

    # if overwrite not requested should fail on reprocessing
    with pytest.raises(OSError):
        converter.from_dx_to_nx(
            input_file=dxfile_path,
            output_file=output_file,
            duplicate_data=duplicate_data,
            overwrite=False,
        )

    # if overwrite requested should succeed
    converter.from_dx_to_nx(
        input_file=dxfile_path,
        output_file=output_file,
        overwrite=True,
        duplicate_data=duplicate_data,
    )



import numpy
import logging
from silx.utils.enum import Enum as _Enum
from silx.io.utils import open as open_hdf5
from silx.io.utils import get_data

from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths

_logger = logging.getLogger(__name__)


__all__ = [
    "OUPUT_CHECK",
    "compare_volumes",
]


class OUPUT_CHECK(_Enum):
    COMPARE_VOLUME = "compare-output-volume"


def compare_volumes(edf_volume_as_urls: tuple, hdf5_scan: NXtomoScan) -> tuple:
    """
    :param tuple edf_volume_as_urls: ordered list of the expected volume.
                                     contains (DataUrl, bool). first element is the location of the orginal frame,
                                     second is a boolean notifying if the frame has been modified or not.
                                     If yes then we cannot do the comparaison.
    :param NXtomoScan hdf5_scan: NXtomo that we want to check the construction.

    On this function we compare a final volume and it's construction. The idea here is more to prevent from
    some 'external' issues such as a file behing not accessible or some 'rock in the shoe'
    """
    issues = set()
    with open_hdf5(hdf5_scan.master_file) as h5f:
        entry_node = h5f[hdf5_scan.entry]
        nexus_path_version = entry_node.attrs.get("version", None)
        nexus_paths = get_nexus_paths(nexus_path_version)
        detector_dataset = entry_node[nexus_paths.PROJ_PATH]

        for i_frame, ((url_original, modified), output_frame) in enumerate(
            zip(edf_volume_as_urls, detector_dataset)
        ):
            original_data = get_data(url_original)
            if modified:
                _logger.info(
                    f"skip comparaison of frame {i_frame}, expected to be different"
                )
            elif original_data.dtype != detector_dataset.dtype:
                # TODO: in this case maybe we want to be more 'accommodating'
                issues.add(
                    f"orignial data and new dataset have different data type ({original_data.dtype} vs {detector_dataset.dtype})"
                )
            elif not numpy.allclose(original_data, output_frame):
                issues.add(
                    f"difference found on frame  {i_frame}. Orignal url is {url_original.path()}"
                )
    return tuple(issues)



# coding: utf-8

from .edfconfig import TomoEDFConfig, generate_default_edf_config  # noqa F401,F403
from .hdf5config import (  # noqa F401,F403
    TomoHDF5Config,
    generate_default_h5_config,
)

from .dxconfig import DXFileConfiguration  # noqa F401,F403

from .fluoconfig import TomoFluoConfig, generate_default_fluo_config  # noqa F401,F403



# coding: utf-8

"""
contains the HDF5ConfigHandler
"""

from __future__ import annotations

import logging
import os
from nxtomomill.utils._os import abspath

from nxtomomill.converter.hdf5.utils import get_default_output_file

from .hdf5config import TomoHDF5Config

_logger = logging.getLogger(__name__)

__all__ = [
    "SETTABLE_PARAMETERS_UNITS",
    "SETTABLE_PARAMETERS_TYPE",
    "SETTABLE_PARAMETERS",
    "TomoHDF5ConfigHandler",
]

SETTABLE_PARAMETERS_UNITS = {
    "energy": "kev",
    "x_pixel_size": "m",
    "y_pixel_size": "m",
    "detector_sample_distance": "m",
}

SETTABLE_PARAMETERS_TYPE = {
    "energy": float,
    "x_pixel_size": float,
    "y_pixel_size": float,
    "detector_sample_distance": float,
}

SETTABLE_PARAMETERS = SETTABLE_PARAMETERS_UNITS.keys()


def _extract_param_value(key_values):
    """extract all the key / values elements from the str_list. Expected
    format is `param_1_name param_1_value param_2_name param_2_value ...`

    :param str_list: raw input string as `param_1_name param_1_value
                         param_2_name param_2_value ...`
    :return: dict of tuple (param_name, param_value)
    """
    if len(key_values) % 2 != 0:
        raise ValueError(
            "Expect a pair `param_name, param_value` for each " "parameters"
        )

    def pairwise(it):
        it = iter(it)
        while True:
            try:
                yield next(it), next(it)
            except StopIteration:
                # no more elements in the iterator
                return

    res = {}
    for name, value in pairwise(key_values):
        if name not in SETTABLE_PARAMETERS:
            raise ValueError(f"parameters {name} is not managed")
        if name in SETTABLE_PARAMETERS_TYPE:
            type_ = SETTABLE_PARAMETERS_TYPE[name]
            if type_ is not None:
                res[name] = type_(value)
                continue
        res[name] = value
    return res


class TomoHDF5ConfigHandler:
    """
    Create a TomoHDF5Config from argparse option provided from the CLI.

    ..warning:: If a configuration file (--config_file) is given then other argparse options will be ignored.
    """

    def __init__(self, argparse_options):
        self._argparse_options = argparse_options
        self._config = None
        self.build_configuration()

    @property
    def configuration(self) -> TomoHDF5Config | None:
        return self._config

    @property
    def argparse_options(self):
        return self._argparse_options

    def build_configuration(self):
        if self.argparse_options.config_file is not None:
            assert isinstance(
                self.argparse_options.config_file, str
            ), f"{self.argparse_options.config_file}, {type(self.argparse_options.config_file)}"
            # from the configuration file
            _logger.warning(
                "A configuration has been given. All other options at the exception of the input and output file will be ignored"
            )
            config = TomoHDF5Config.from_cfg_file(self.argparse_options.config_file)
            if self.argparse_options.input_file is not None:
                config.input_file = self.argparse_options.input_file
            if self.argparse_options.output_file is not None:
                config.output_file = self.argparse_options.output_file
        else:
            # from argparse
            config = TomoHDF5Config()

        options = self.argparse_options
        # check input and output file

        # propagate options defined by the user to the config file.
        # policy: all options when not defined have a default value set to None
        # So we don't have to set twice the default value from the source code (class and argparse)
        for opt_name in (
            "input_file",
            "output_file",
            "file_extension",
            "single_file",
            "no_master_file",
            "overwrite",
            "entries",
            "sub_entries_to_ignore",
            "raises_error",
            "field_of_view",
            "request_input",
            "default_data_copy",
            "sample_x_keys",
            "sample_y_keys",
            "translation_z_keys",
            "sample_detector_distance_keys",
            "valid_camera_names",
            "rotation_angle_keys",
            "exposure_time_keys",
            "sample_x_pixel_size_keys",
            "sample_y_pixel_size_keys",
            "init_titles",
            "zseries_init_titles",
            "multitomo_init_titles",
            "back_and_forth_init_titles",
            "dark_titles",
            "flat_titles",
            "projection_titles",
            "alignment_titles",
        ):
            opt_value = getattr(options, opt_name)
            if opt_value is not None:
                setattr(config, opt_name, opt_value)

        # handle specific use cases
        if options.debug is not None:
            config.log_level = logging.DEBUG

        if options.set_params is not None:
            extra_params = _extract_param_value(options.set_params)
            for param, param_value in extra_params.items():
                setattr(config, param, param_value)

        if config.input_file is None:
            raise ValueError("No input file provided")

        if not config.output_file:
            # Warning: output_file can be None or "" at this stage.
            if config.file_extension is None:
                raise ValueError(
                    "If no output file provided you should provide the extension"
                )

            config.output_file = get_default_output_file(
                input_file=config.input_file,
                output_file_extension=config.file_extension.value,  # pylint: disable=E1101
            )
        elif os.path.isdir(config.output_file):
            # if the output file is only a directory: consider we want the same file basename with default extension
            # in this directory
            input_file, _ = os.path.splitext(config.input_file)
            config.output_file = os.path.join(
                abspath(config.output_file),
                os.path.basename(input_file)
                + config.file_extension.value,  # pylint: disable=E1101
            )

        self._config = config
